"""Interface to read tabular data stored in an Excel 2007- (xlsx) workbook file.

*** Version History:

2019-07-23, first functional version
2019-07-27, change to subclass of Table
"""
# **** general test for unique subjectID here, or in calling module ??? *********

# handle record as dict for simpler recode_fcn and accept_row  ??? *******

# ***** define columns by column address or also by column index or column header ??? **********

from openpyxl import load_workbook
from openpyxl.utils.exceptions import InvalidFileException
from openpyxl.utils import column_index_from_string
from openpyxl.utils.cell import get_column_letter


from ItemResponseCalc import table_reader

# import logging
# logger = logging.getLogger(__name__)


class FileFormatError(RuntimeError):
    """Format error causing non-usable data"""

#
# class ParameterError(pc_base.FileReadError):
#     """Error in calling parameters causing non-usable data"""


class ArgumentError(RuntimeError):
    """Error in calling parameters causing non-usable data"""


# ------------------------------------------------------------------------

class Table(table_reader.Table):
    """Interface to ONE sheet of an Excel xlsx workbook storing tabular data
    """
    def __init__(self, file_path,
                 fields=None,
                 sheet=None,
                 header_row=1,
                 **kwargs):
        """
        :param file_path: Path to file for reading
        :param fields: list of fields to be returned in each record
            if None, ALL fields are returned
        :param sheet: (optional) sheet name.
            If undefined, only the first work-book sheet is used.
        :param header_row: row with header labels for all columns
            if header_row is None or == 0, column addresses are used as field names
        :param kwargs: additional properties for super-class,
            field_types, keyed, accept_row, missing_values, recode_fcns
        """
        self.sheet = sheet
        if header_row is None:
            header_row = 0
        self.header_row = header_row
        super().__init__(table_id=file_path, fields=fields, **kwargs)

    def preview(self, max_records=10, keyed=False):
        """Read headers and some raw data rows from the file,
        as stored without any type casting or other checks
        :param max_records: limit on the number of records
        :param keyed: boolean: if True, return data as dict # *********************** ???
        :return: iterator over data rows, as stored without any type casting or other checks
        """
        wb = self._get_workbook()
        ws = self._get_sheet(wb)
        # result = []
        rows = ws.iter_rows(min_row=1)
        for (row_n, row) in enumerate(rows):
            if row_n > max_records:
                break
            # result.append([cell.value for cell in row])
            yield [cell.value for cell in row]
        wb.close()
        # return result

    def __iter__(self):
        """Open file for read-only access,
        yielding one row at a time,
        including only selected fields,
        selected, recoded, and pre-processed by self.process_row().
        :return: iterator object over file rows,
            starting just after self.header_row
        """
        wb = self._get_workbook()
        ws = self._get_sheet(wb)
        header = self._get_header(ws)
        if self.fields is None:
            fields_index = None
        else:
            fields_index = [header.index(f) for f in self.fields]
            # ************** must catch ValueError if f is not in header
        rows = ws.iter_rows(min_row=self.header_row + 1)
        for (row_n, row) in enumerate(rows):
            if row_n % self.sample_factor == 0:
                if fields_index is None:
                    row_val = [r.value for r in row]
                else:
                    row_val =[row[i].value for i in fields_index]
                row_val = self.process_row(row_val)
                if row_val is not None:
                    yield row_val
        wb.close()  # ******** might never get here ******** need context manager ?

    def __len__(self):
        """Number of rows in file, except header
        :return: scalar integer
        """
        wb = self._get_workbook()
        ws = self._get_sheet(wb)
        n = ws.max_row - self.header_row
        wb.close()
        return n // self.sample_factor

    # ------------------------------------------------ Private:
    def _get_workbook(self):
        """Open work-book and get ONE selected sheet
        :return: ws = a work-sheet object
        """
        try:
            return load_workbook(str(self.table_id), read_only=True)
        except InvalidFileException:
            raise FileFormatError(f'Cannot load workbook from file {self.table_id.stem}')

    def _get_sheet(self, wb):
        """Get ONE selected sheet from an open workbook
        :param wb: active WorkBook object
        :return: ws = a work-sheet object
        """
        if self.sheet is None:
            self.sheet = wb.sheetnames[0]
        try:
            return wb[self.sheet]
        except IndexError:
            raise FileFormatError(f'Cannot load sheet {repr(self.sheet)} from file {self.table_id.stem}')

    def _get_header(self, ws):
        """get file header from the worksheet, if it has a header row
        :param ws: selected WorkSheet of a Workbook open for read-only access
        :return: list with header name strings from the file
            or column addresses if there is no header row
        """
        if self.header_row == 0:
            row1 = ws[1]
            return [get_column_letter(i) for i in range(1, 1+len(row1))]
        else:
            header_row = ws[self.header_row]
            return [cell.value for cell in header_row]

# --------------------------------------------------- help sub-functions


# -----------------------------------------------------------------------------
# if __name__ == '__main__':
#     from pathlib import Path
#
#     HAQ_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry'
#     # top-dir for everything to be used here
#
#     ioiha_data_path = HAQ_PATH / 'IRT-IOI-HA' / 'IOI-HA-data'
#     # to ioi-ha data sets OTHER THAN the Swedish Quality Registry
#
#     ioiha_hickson = ioiha_data_path / 'Hickson'
#
#     hickson_path = ioiha_hickson / 'Short_Eartrak AUS.xlsx'
#
#     table = Table(file_path=hickson_path,
#                   fields='CDEFGHI',
#                   header_row=0,
#                   missing_values=['.'])
#
#     print('\n*** Using table(header_row=0): \n')
#     print('Number of rows = ', len(table))
#     for (row_n, row) in enumerate(table):
#         if row_n > 11:
#             break
#         print(f'row #{row_n} = {row}')
#
#     print('table.preview(10)=\n', table.preview(10))
#     for row in table.preview(10):
#         print(row)
#
#     table = Table(file_path=hickson_path,
#                   fields=['Q01', 'Q02', 'Q03', 'Q04', 'Q05', 'Q06', 'Q07'],
#                   header_row=1,
#                   missing_values=['.'])
#
#     print('\n*** Using table(header_row=1): \n')
#     print('Number of rows = ', len(table))
#     for (row_n, row) in enumerate(table):
#         if row_n > 10:
#             break
#         print(f'row #{row_n} = {row}')
#
    # print('Hickson file.preview(10)=\n',table.preview(10))




"""This module implements a Bayesian model of subjects' responses
to items in a questionnaire, using the Graded Response Model of Item Response Theory.

*** Main Classes:

OrdinalItemResponseModel, defining a complete Bayesian model with all parameters
    adapted to a data set with individual responses to each item in a questionnaire.
    obtained from subjects in one or more different groups.

*** Usage Example: see template script run_irt.py


*** Reference:
Leijon, Kramer et al.:
Analysis of Data from the International Outcome Inventory for Hearing Aids (IOI-HA)
using Bayesian Item Response Theory. 2019, Manuscript in preparation.

Arne Leijon, 2019-07-08, first functional version
2019-07-23, set Initial Trait Scale only here,
    for use by both item_scale and item_respondent initialization
2019-08-15, trying new initialization method, no boolean item_trait_map input
2019-08-16, max_hours limit for learn
2019-08-21, using Pool multi-processing for learn
"""
# **** allow string item_id ? *********************
# **** allow string trait_id ? *********************

import numpy as np
import datetime as dt

from collections import Counter

from .item_respondents import ResponseGroup
from .item_trait_selector import TraitProb
from .item_scale import make_scales
from .trait_precision import prior_precision_within, prior_precision_among
from .ir_predictive import GaussianGivenParam

from multiprocessing.pool import Pool
import os

import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.INFO)
# logger.setLevel(logging.DEBUG)  # test

# ----------------------------------------------
__version__ = '2020-06-06'

INITIAL_TRAIT_SCALE = 3.  # larger? ***********
# = crude assumed initial scale of model trait values
# ----------------------------------------------


usePool = True
# usePool = False # for TEST


def _pool_size(n):
    """Estimate number of Pool sub-processes
    :param n: total number of independent jobs to share between processes
    :return: n_processes
    """
    # NOTE: cpu_count() returns n of LOGICAL cpu cores.
    return min(os.cpu_count(), n)


# ---------------------------------------------- Main class:
class OrdinalItemResponseModel:
    """A Bayesian IRT model for discrete ORDINAL rating data,
    available as INDIVIDUAL results in a 2D array R with elements
    R[s, i] = ordinal integer-coded response by s-th subject to the i-th item.

    The number of response alternatives may differ among items.
    The i-th item has L_i ordinal integer response values.

    NOTE: In most modules of this package,
    the math notation assumes that R[s, i] is an integer in range(L_i),
    i.e., it can have integer values 0, 1,..., L_i-1.

    However, in the input data files,
    the recorded responses are usually coded as values 1, 2,..., L_i,
    with missing values encoded as zero.
    This response notation is used in module item_response_data.

    All model parameters are treated as random variables.
    The main parameter distributions are represented by equally probable samples.

    This package uses the Graded Response Model (Samejima, 1969, 1997):
    Each rating R_si is assumed to be determined by ONE latent random Trait variable
    Y_st = a real-valued random variable in the range (-inf, inf).
    There may be one or more separate trait variables,
    each determining the responses to one or more items.

    IF the i-th item is associated with the t-th trait,
    the latent variable Y_st is assumed to determine the response R_si as
    a_{i, l} < Y_st <= a_{i, l+1} => R_si = l,
    where the response thresholds are defined in
    an array (a_k0, ..., a_kL) with strictly increasing elements, and
    a_k0 = -inf; a_kL = +inf

    The latent random variable Y_st has a logistic distribution
    with location theta_st = a subject-specific TRAIT variable,
    for the t-th trait of the s-th subject.

    The trait (location) parameters represent
    the individual performance measures to be estimated.

    The response thresholds are represented by
    an item_scale.OrdinalItemScale object for each item,
    adapted to the distribution of responses from all subjects.

    Each item response is modelled as determined by ONE trait variable,
    identified by an item_trait_selector.TraitSelector object
    connected as a property of each item scale object.
    The association between traits and items is automatically
    estimated by the learning procedure.

    The scale of model parameters is arbitrary, assigned as follows:
    The scale of the latent variable is unity during model estimation.
    However, after learning, all model parameters may be re-scaled jointly,
    such that the predicted global variance is unity for all traits.

    The model is implemented by main properties
    groups = a list with one item_respondents.RespondentGroup instance for each group of subjects.
        Each group model includes all responses for each subject.
    scales = a list of item_scale.OrdinalItemScale objects,
        each defining rating intervals for the mapping
        from latent interval-scale values to ordinal ratings.
        Each scale object has its own item_trait_selector.TraitSelector object,
        defining the association between traits and the item.
    trait_prob = an item_trait_selector.TraitProb object,
        representing the overall probability for each trait to be used for each item.
    precision_within = a trait_precision.TraitPrecisionWithinGroup object, representing
        a Wishart distribution of the intra-group precision of trait vectors theta.
    precision_among = a trait_precision.TraitPrecisionAmongGroups object, representing
        a gamma distribution for the precision of mean trait vectors between groups.
    """
    @classmethod
    def initialize(cls, data_source,
                   n_traits=None,
                   n_scale_samples=1,
                   n_subject_samples=50,
                   trait_scale=INITIAL_TRAIT_SCALE):
        """Initialize a model crudely from given item response data.
        :param data_source: a single ItemResponseDataSet instance,
            including response data from one or more groups of subjects.
        :param n_traits: (optional) scalar integer max number of effective latent traits
        :param n_scale_samples: (optional) number of samples of scale parameters tau for each item
        :param n_subject_samples: (optional) number of samples of trait parameter theta for each subject
        :param trait_scale: (optional) scalar initially assumed scale of model trait values
        :return: a cls instance, crudely initialized for given data_source
            to be refined later, by method learn

        Method:
        1: Use item response counts to crudely initialize all item scales
        2.1: Initialize groups with individual response data for all items, using scales
        2.2: Use PCA or Factor Analysis (with varimax?) on these item-specific trait variables
        2.3: Transform all item-specific traits to a space with n_traits dimensions.
        3: Initialize precision_within, using data from these groups
        4: Initialize groups with only summary data
        5: Initialize precision_among using data from all groups
        """
        assert len(data_source.groups) > 0, 'Must have at least one group'

        item_response_count = data_source.item_response_count
        # item_response_count[i] = 1D array of response counts for i-th item
        n_items = len(item_response_count)
        if n_traits is None:
            n_traits = n_items

        # ----------------------------------------- initialize scales
        trait_prob = TraitProb.initialize(n_traits)
        logger.debug(f'Prior TraitProb.alpha = {np.array2string(trait_prob.alpha, precision=4)}')
        scales = make_scales(item_response_count,
                             trait_prob,
                             n_scale_samples,
                             trait_scale)
        logger.debug('\n' +
                     '\n'.join(f'Init scales[{i}] mean tau: '
                               + np.array2string(np.mean(sc.tau, axis=0), precision=3)
                               + '; trait_sel.mean= '
                               + np.array2string(sc.trait_sel.mean, precision=2)
                               for (i, sc) in enumerate(scales)))  # ******** do this in scales ********

        # ----------------------------------------- initialize groups with subjects
        groups = [ResponseGroup.initialize_by_item(subject_responses=g_responses,
                                                   scales=scales,
                                                   trait_scale=trait_scale,
                                                   n_samples=n_subject_samples,
                                                   name=g_name)
                  for (g_name, g_responses) in data_source.groups.items()]
        # = groups with individual response data for each item
        # g_responses = iterable yielding subject responses, coded as a 1D list with
        # g_responses[s][i] = integer index of response to i-th item
        trait_cov = np.sum(g.cov_all()
                           for g in groups)
        transform_matrix = _trait_rotation(trait_cov, n_traits)
        for g in groups:
            g.transform_traits(transform_matrix)

        # ----------------------------------------- initialize precision_within
        precision_within = prior_precision_within(n_traits)
        precision_within.adapt((g.mean_cov_theta for g in groups))

        # ----------------------------------------- initialize precision_among
        precision_among = prior_precision_among(n_traits)
        logger.debug(f'prior_precision_among.b= {np.mean(precision_among.b)}')

        return cls(data_source.questionnaire,
                   groups, scales, trait_prob,
                   precision_within, precision_among)

    # --------------------------------------------------------------------
    def __init__(self, questionnaire,
                 groups, scales, trait_prob,
                 precision_within, precision_among):
        """
        :param questionnaire: Questionnaire object with info about the items
        :param groups: list with ResponseGroup elements
        :param scales: list of OrdinalItemScale instances, one for each item
        :param trait_prob: single TraitProb object, prior for all scale.trait_sel objects
        :param precision_within: Wishart distribution object
            representing inter-individual precision matrix
            for subject trait vectors, assumed valid for all subject groups
        :param precision_among: gamma distribution object
            representing inter-group precision (inverse variance)
            for mean trait vectors across groups.
        """
        self.version = __version__
        self.questionnaire = questionnaire
        self.groups = groups
        self.scales = scales
        self.trait_prob = trait_prob
        # self.trait_item_map = trait_item_map
        self.precision_within = precision_within
        self.precision_among = precision_among
        # self.latent_scale = 1.  # scale parameter of logistic latent variable
        # may be changed after learning, by method standardize
        self.log_prob = []

    def __repr__(self):  # **********************************
        return (self.__class__.__name__ + '('
                + f'\n\t questionnaire= {self.questionnaire.__class__.__name__} '
                + f'object with {self.questionnaire.n_items} items,'
                + '\n\t groups= [\n\t\t'
                + ',\n\t\t'.join(f'{g.__class__.__name__}(name={repr(g.name)}, with {g.n_subjects} subjects)'
                                 for g in self.groups) + '],'
                + f'\n\t scales= [{len(self.scales)} item-scale objects],'
                + f'\n\t trait_prob = {self.trait_prob.__class__.__name__} object, '
                + f'mapping items to {self.n_traits} latent traits, '
                + f'\n\t precision_within= {self.precision_within.__class__.__name__} object,'
                + f'\n\t precision_among= {self.precision_among.__class__.__name__} object,'
                + f'\n\t version= {self.version}'
                + ')')

    @property
    def n_groups(self):
        return len(self.groups)

    @property
    def n_subjects(self):
        return sum(g.n_subjects for g in self.groups)

    @property
    def n_items(self):
        return len(self.scales)

    @property
    def n_traits(self):
        return len(self.trait_prob.mean)

    # ------------------------------------------ General VI learn algorithm:
    def learn(self,
              min_iter=10,
              min_step=0.01,
              max_iter=100,
              max_hours=10,
              callback=None):
        """Learn from all observed response data stored in self.groups,
        using Variational Inference (VI).

        This method adapts sampled distributions for all model parameters
        to maximize a lower bound to the total likelihood of the observed data.
        The resulting sequence of lower-bound values is theoretically guaranteed to be non-decreasing,
        but the sampling variability may cause small variations around the non-decreasing trend.
        :param min_iter: (optional) minimum number of learning iterations
        :param min_step: (optional) minimum data log-likelihood improvement,
                 over the latest min_iter iterations,
                 for learning iterations to continue.
        :param max_iter: (optional) maximum number of iterations, regardless of result.
        :param max_hours = (optional) maximal allowed running time, regardless of result.
        :param callback: (optional) function to be called after each iteration step.
            If callable, called as callback(self, log_prob)
            where log_prob == scalar last achieved value of VI lower bound
        :return: log_prob = list of log-likelihood values, one for each iteration.
        Result: updated properties of self
        """
        min_iter = np.max([min_iter, 1])
        end_time = dt.datetime.now() + dt.timedelta(hours=max_hours)
        # last allowed time to start new VI iteration
        log_prob = []
        while (len(log_prob) <= min_iter
               or (log_prob[-1] - log_prob[-1 - min_iter] > min_step
                   and (len(log_prob) < max_iter)
                   and (dt.datetime.now() < end_time))):
            log_prob.append(self.adapt())
            if callable(callback):
                callback(self, log_prob[-1])
            logger.info(f'Done {len(log_prob)} iterations. LL={log_prob[-1]:.1f}')
        self.log_prob = np.array(log_prob)
        return self.log_prob

    def adapt(self):
        """One adaptation step for all parameters
        using all response data stored in self.groups.
        :return: ll = scalar lower bound to data log-likelihood,
            incl. negative contributions for KLdiv re priors
        """
        ll = 0.
        ll += self._adapt_prec()
        ll += self._adapt_scales()
        ll += self._adapt_groups()
        # *** should be done in this order,
        # because prec are used in scales update
        # and prec and scales are used in groups update.
        # All KLdiv calculations must use SAME distributions
        # i.e., the current value AFTER adaptations  *********************** CHECK !!!
        return ll

    def _adapt_scales(self):
        """One adaptation step for all parameters
        using all response data stored in self.groups.
        :return: ll = scalar lower bound to data log-likelihood
        """
        _adapt_args = (self.groups, self.trait_prob)
        if usePool:
            n_pool = _pool_size(len(self.scales))
            ch_size = 1  # *** probably slightly faster than ch_size=2
            logger.debug(f'_adapt_scales Pool: n_pool= {n_pool}, ch_size= {ch_size}')
            with Pool(n_pool) as p:
                self.scales = list(p.imap(_adapt, ((sc, _adapt_args)
                                                   for sc in self.scales),
                                          chunksize=ch_size))
        else:
            self.scales = list(map(_adapt, ((sc, _adapt_args)
                                            for sc in self.scales)))

        for (i, sc) in enumerate(self.scales):
            logger.debug(f'scales[{i}] mean tau: '
                         + np.array2string(np.mean(sc.tau, axis=0),
                                           precision=2, suppress_small=True)
                         + ' trait_sel.mean= '
                         + np.array2string(sc.trait_sel.mean,
                                           precision=4, suppress_small=True))
        sum_r = sum(sc.trait_sel.mean
                    for sc in self.scales)
        logger.debug('TraitProb sum responsibility = '
                     + np.array2string(sum_r, precision=2, suppress_small=True))
        ll = self.trait_prob.adapt(sum_r)
        logger.debug(f'TraitProb.adapted: mean='
                     + np.array2string(self.trait_prob.mean, precision=3, suppress_small=True)
                     + f'; ll= -KLdiv = {ll}')
        # NOTE: scale.trait_sel kl_div must be calculated AFTER TraitProb update
        for sc in self.scales:
            ll -= sc.relative_entropy_re_prior(self.trait_prob)  # incl. both eta and trait_sel
        logger.debug(f'scales ll = {ll:.1f}')
        return ll

    def _adapt_groups(self):
        """One adaptation step for all parameters
        using all response data stored in self.groups.
        :return: ll = scalar lower bound to data log-likelihood
        """
        adapt_args = (self.scales, self.precision_within, self.precision_among)
        if usePool:
            n_pool = _pool_size(len(self.groups))
            ch_size = 1
            logger.debug(f'_adapt_groups Pool: n_pool= {n_pool}, ch_size= {ch_size}')
            with Pool(n_pool) as p:
                self.groups = list(p.imap(_adapt, ((g, adapt_args)
                                                   for g in self.groups),
                                          chunksize=ch_size))
        else:
            logger.debug(f'_adapt_groups map, NoPool')
            self.groups = list(map(_adapt, ((g, adapt_args)
                                            for g in self.groups)))

        for g in self.groups:
            logger.debug(f'group {repr(g.name)}: mu.loc= '
                         + np.array2string(g.mu.loc, precision=3, suppress_small=True)
                         + f'; ll={g.ll:.2f}')
        ll = sum(g.ll for g in self.groups)
        return ll

    def _adapt_prec(self):
        """One adaptation step for all parameters
        using all response data stored in self.groups.
        :return: ll = scalar lower bound to data log-likelihood
        """
        ll = self.precision_within.adapt((g.mean_cov_theta
                                          for g in self.groups))
        logger.debug('within-groups var= '
                     + np.array2string(np.diag(self.precision_within.mean_inv),
                                       precision=3))
        ll += self.precision_among.adapt([g.mu.mean2
                                          for g in self.groups])
        logger.debug('among-groups var= '
                     + np.array2string(self.precision_among.mean_inv,
                                       precision=3))
        logger.debug(f'prec ll = {ll:.1f}')
        return ll

    # ------------------------------------------------------- Result Display Functions:

    def item_trait_map(self):
        """Mapping with ONE trait corresponding to each item
        :return: itp = 2D boolean array
            itp[i, t] == True <=> i-th item responses were determined by t-th trait
        """
        return np.array([sc.trait_sel.mean > 0.5 for sc in self.scales])

    def prune(self):
        """Reduce model complexity by deleting any latent trait variables
        that did not correspond to any item
        :return: None
        Result: modified internal model properties
        """
        it_map = self.item_trait_map()
        keep = np.any(it_map, axis=0)
        self.trait_prob.prune(keep)
        self.precision_within.prune(keep)
        self.precision_among.prune(keep)
        for sc in self.scales:
            sc.prune(keep)
        for g in self.groups:
            g.prune(keep)

    def standardize(self):
        """Rescale all model parameters, after learning, such that
        predictive individual trait variance is unity for all traits
        :return: None
        """
        s = np.sqrt(self.predictive_individual_var())
        logger.info(f'Standardizing by factors '
                     + np.array2string(s, precision=3))
        for g in self.groups:
            g.standardize(s)
        self._adapt_prec()
        for sc in self.scales:
            sc.standardize(s)

    # ------------------------------------------------------ Raw Descriptive Data:
    def item_response_count(self):
        """Total frequency of responses at each ordinal level, for each item,
        by subjects in each included group.
        :return: dict with elements (group_key, g_count), where
            g_count is a list of collections.Counter objects,
            g_count[i][r] = number of responses == r for i-th item,
            with r==0 indicating missing response, and
            r = 1, ..., L_i are actual responses.
        """
        return {g.name: [Counter(sr_i)
                         for sr_i in g.subject_responses]
                for g in self.groups}

    # ------------------------------------------------------ Predictive Models:
    def predictive_individual_var(self):
        return np.diag(self.predictive_individual_cov())

    def predictive_individual_cov(self):
        """Covariance of the zero-mean Predictive distribution of IRT traits
        for a RANDOM INDIVIDUAL in any not yet seen future population,
        similar to the sub-populations represented by the included subject groups.
        :return: 2D array with covariance matrix

        Method: The exact predictive distribution is a continuous mixture
        of Student-t distributions.
        Samples from the exact distribution can be generated,
        but it may be sufficient to know the exact covariance
        derived from self.precision_within and self.precision_among
        """
        n_g = self.n_groups
        if n_g <= 1:
            logger.warning(f'Uncertain predictive distribution with only {n_g} groups')
        return (self.precision_within.mean_inv
                + self.precision_among.mean_inv * np.eye(self.n_traits))

    # def predictive_mean(self):
    #     """Predictive distribution of MEAN IRT traits
    #     in any not yet seen future population, similar to any of the sub-populations
    #     represented by the included groups of subjects.
    #     :return: PredictiveTrait instance
    #     """
    #     return GaussianGivenParam(mu=np.zeros(self.n_traits),
    #                               tau=self.precision_among)

    def predictive_individual(self):
        """Sampled Predictive distribution of IRT traits for a RANDOM INDIVIDUAL
        in ANY not yet seen future population, similar to any of the sub-populations
        represented by the included groups of subjects.
        :return: single ir_predictive.GaussianGivenParam object

        Method: the mean of the distribution is a zero-mean Gaussian random variable,
            with precision = estimated precision among groups.
        """
        return GaussianGivenParam(mu=GaussianGivenParam(mu=np.zeros(self.n_traits),
                                                        tau=self.precision_among),
                                  tau=self.precision_within)

    # ---------------------------------- Descriptive statistics:

    def mean_subject_ratings(self):
        """Mean raw ratings, across items, for each respondent in each group
        NOTE: Only informative: The raw ratings NOT on an interval scale!
        :return: dict with elements (group_key, mean_rating)
            mean_rating[s] = mean rating across items, by s-th subject
        """
        mr = dict()  # space for result
        for g in self.groups:
            try:
                mr[g.name] = g.mean_response()
            except AttributeError:
                pass  # no-name group, should not happen
        return mr

    def mean_scaled_subject_ratings(self):
        """Mean ratings, across items, for each respondent in each group,
        with raw ratings re-encoded on point-estimated interval scale.
        NOTE: Only informative, point-estimates are still discrete.
        :return: dict with elements (group_key, mean_rating)
            mean_rating[s] = mean rating across items, by s-th subject
        """
        s = [np.mean(sc.typical_trait(), axis=0)
             for sc in self.scales]
        # s = point-estimated scale values on common interval scale with mean == 0
        # s[i][r] = value for r-th response to i-th item
        return {g.name: np.mean(g.scaled_subject_responses(s), axis=0)
                for g in self.groups}

    def mean_subject_traits(self):
        """Mean latent trait variables in interval-scale domain, range (-inf, +inf),
        for each subject group,
        with equal weight for each item,
        i.e., higher weight if the trait is controlling responses to several items.
        This allows a fair comparison with mean_subject_rating,
        but the trait averaging presumes that separate traits (and items) represent
        measures that make sense to project onto a common uni-dimensional scale.
        :return: dict with elements (group_key, mean_trait), where
            mean_trait = 1D array with individual mean trait values
            mean_trait[s] = mean trait for s-th subject, weighted by item_trait_map
        """
        w = np.sum([sc.trait_sel.mean
                    for sc in self.scales], axis=0)
        w /= np.sum(w)
        # = normalized weight, for fair comparison to raw average rating across items
        return {g.name: np.mean(np.dot(g.theta, w), axis=0)
                for g in self.groups}

    def estim_var_mean_subject_traits(self):
        """Estimation-variance of mean_subject_traits() results,
        used to calculate scale error component of conditional variance given mean rating.
        :return: dict with elements (group_key, estim_var), where
            estim_var = scalar variance estimate for the group
        """
        w = np.sum([sc.trait_sel.mean
                    for sc in self.scales], axis=0)
        w /= np.sum(w)
        # = normalized weight, for fair comparison to raw average rating across items
        # for same weighting as in mean_subject_traits()
        # divide sample variance by n_samples to estimate variance of sample mean
        return {g.name: np.mean(np.var(np.dot(g.theta, w), axis=0)) / g.n_samples
                for g in self.groups}


# ------------------------------ module-internal help functions
def _trait_rotation(c, n_traits):
    """Transformation matrix for trait vectors to (sub-)space
    defined by eigenvectors of cov matrix
    :param c: 2D array with covariance matrix for all initial item-defined traits
        summed across all subjects in all groups
        c.shape == (n_items, n_items)
    :return: proj: 2D projection matrix with orthogonal COLUMN vectors
        proj[:, t] = unit direction vector for t-th trait
        proj.shape == (n_items, n_traits)
        where n_traits <= n_items
        The vectors are scaled such that transformation preserves trait variance
        in the new coordinate system,
        so the initial item scale thresholds should still be reasonably valid.
    """
    (e_val, e_vec) = np.linalg.eigh(c)
    mean_var = np.mean(np.diag(c))
    s = np.sign(e_vec[np.argmax(np.abs(e_vec), axis=0), np.arange(e_vec.shape[1])])
    e_vec *= s
    e_det = np.linalg.det(e_vec)
    proj = e_vec[:, ::-1][:, :n_traits]  # in decreasing e_val order
    new_var = e_val[::-1][:n_traits]  # in same order
    s = np.sqrt(mean_var / new_var)
    # = scale factor to preserve trait variance
    proj *= s
    c_proj = proj.T @ c @ proj  # *** for test, should be diagonal
    return proj
# ---------------------------------------------------------------------


# ------------------------------ help function for Pool multi-tasking:
# used by map or Pool().imap:
def _adapt(task):
    """dispatch call to given object.adapt(...)
    :param task: arguments defining the task:
        task[0] = object whose adapt method is called
        task[1] = positional arguments for adapt method
    :return: returned object from the called adapt method,
        i.e., an updated copy of object task[0]
    """
    # obj = task[0]
    # arg = task[1]
    # logger.debug(f'{obj}.adapt({arg})')
    return task[0].adapt(*task[1])


# ------------------------------ help functions for predictive calculations:



"""set up logging formats for ItemResponseCalc package"""

import logging


def setup(log_file):
    """create logging handlers
    :param log_file: string file name for logger FileHandler
    :return: None
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    # root_logger.setLevel(logging.DEBUG)
    # NOTE: haq_model.usePool MUST be False, to get desired logging level in Pool-ed sub-modules

    # *** create handlers explicitly, to specify encoding:
    if log_file is not None:
        fh = logging.FileHandler(log_file, mode='w', encoding='utf-8')
        fh.setFormatter(logging.Formatter('{asctime} {name} {levelname}: {message}',
                                          style='{',
                                          datefmt='%Y-%m-%d %H:%M:%S')
                        )
        root_logger.addHandler(fh)

    # --------------------------------- No console log if running on HCALC?
    # NO, just keep the console log anyway; check nohup.out for progress log
    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter('{asctime} {name}: {message}',
                                           style='{',
                                           datefmt='%H:%M:%S'))
    root_logger.addHandler(console)


"""This module defines classes for
intra-group and inter-group precision of trait parameters
in an OrdinalItemResponseModel.

The class instances are similar to frozen scipy.stats distribution objects,
but extend functionality by allowing the objects to adapt to observed data.

*** Main Classes:

TraitPrecisionWithinGroup -- Wishart distribution of full precision matrix,
    i.e., inverse of covariance matrix, of individual trait vectors.
    **** might be diagonal, if we require latent traits to be orthogonal ****
    **** but why should they have to be orthogonal ??? *******

TraitPrecisionAmongGroups -- gamma distribution of diagonal precision (inverse variance)
    for the group trait mean variations among groups.
"""
import numpy as np
from scipy.special import multigammaln, psi, gammaln
from scipy.stats import wishart

PRIOR_PRECISION_DELTA = 0.001
# Prior df of within-group Wishart distribution is
# df = D - 1 + PRIOR_PRECISION_DELTA, where
# df = approximate weight of prior precision re. one individual subject

PRIOR_SCALE_WITHIN = 1.
# = approximate scale of individual trait precision within groups

PRIOR_A_AMONG = 0.001
PRIOR_B_AMONG = 0.001
# = prior a, b parameters for prior_precision_among


# ---------------------------------------------- prior precision objects:
def prior_precision_among(n_traits):
    """
    :param n_traits: number of model traits
    :return: prior TraitPrecisionAmongGroups object
    """
    return TraitPrecisionAmongGroups(a=PRIOR_A_AMONG,
                                     b=np.ones(n_traits) * PRIOR_B_AMONG)


def prior_precision_within(n_traits):
    """
    :param n_traits: number of model traits
    :return: prior TraitPrecisionWithinGroup object
    """
    nu = n_traits - 1 + PRIOR_PRECISION_DELTA
    # = degrees-of-freedom to make the prior just barely proper
    # We need a PROPER prior to be able to calculate relative entropy
    # between posterior and prior models.
    return TraitPrecisionWithinGroup(df=nu,
                                     scale=np.eye(n_traits) * PRIOR_SCALE_WITHIN / nu)


# ---------------------------------------------------------------------------
class TraitPrecisionWithinGroup:
    """Wishart distribution of a symmetric precision matrix Lambda of a Gaussian vector
    The probability density function is
    p(Lambda) = (1/C) |Lambda|**((nu - p - 1)/2) exp(- trace(inv(V) Lambda)/2), where
        V = symmetric scale matrix,
        p = length of corresponding Gaussian vector
        nu = degree of freedom parameter >= p,
        Lambda.shape == V.shape == (p, p)
        C = |V|**(nu/2) 2**(nu p / 2) Gamma_p(nu/2) is the normalization factor
    """
    def __init__(self, scale, df=None):
        """
        :param scale: array or array-like symmetric scale matrix V
        """
        self.scale = np.asarray(scale)
        n_traits = len(scale)
        if df is None:
            df = n_traits + PRIOR_PRECISION_DELTA
        self.df = df

    def __repr__(self):
        return self.__class__.__name__ + f'(df= {self.df}, scale= \n {self.scale})'

    @property
    def prior(self):
        return prior_precision_within(self.n_traits)

    @property
    def n_traits(self):
        return len(self.scale)

    @property
    def inv_scale(self):
        return np.linalg.inv(self.scale)

    @property
    def mean(self):
        """E{self}"""
        return self.df * self.scale

    @property
    def mean_inv(self):
        """E{ inv(self) }"""
        if self.df > self.n_traits + 1:
            return self.inv_scale / (self.df - self.n_traits - 1)
        else:
            # mean_inv is undefined
            return np.full_like(self.scale, np.nan)

    @property
    def mode_inv(self):
        """Mode of inv(self) = most probable corresponding covariance matrix
        """
        return self.inv_scale / (self.df + self.n_traits + 1)

    @property
    def mean_log_det(self):
        """E{ ln |self| }
        (self is positive definite, so sign of determinant is always = 1.)
        """
        (s, log_det_scale) = np.linalg.slogdet(self.scale)
        return multi_psi(self.df / 2, self.n_traits) + self.n_traits * np.log(2) + s * log_det_scale

    # def logpdf(self, x):
    #     """ln pdf(x | self), using scipy.stats.wishart
    #     :param x: 2D array or array-like list
    #     :return: lp = scalar lp = ln pdf(x | self)
    #
    #     NOTE: scipy.stats.wishart can do logprob for matrices staced along 3rd axis
    #     but we do not use it.
    #     """
    #     return wishart.logpdf(x, df=self.df, scale=self.scale)

    def rvs(self, size=1):
        """Samples from Wishart distribution
        :param size: scalar integer number of sample matrices
        :return: tau = 3D array
            tau[..., :, :] = ...-th random precision matrix
            tau.shape == (*size, *self.scale.shape)
        """
        return wishart.rvs(df=self.df, scale=self.scale, size=size)

    # ------------------------------------------- MAP learning:
    def adapt(self, group_cov):
        """Update distribution parameters using observed data and prior.
        :param group_cov: iterable of lists of covariance-matrix samples
            for vectors assumed to have precision matrix == self.
            group_cov[g][s] = 2D array of mean current estimated cov for s-th subject in g-th group
        :return: - KLdiv(self || self.prior)
        Result: updated internal parameters of self
        """
        prior = self.prior
        n_s = 0
        cov = np.zeros((self.n_traits, self.n_traits))
        for c in group_cov:
            n_s += len(c)
            cov += np.sum(c, axis=0)
        inv_scale = prior.inv_scale + cov
        self.scale = np.linalg.inv(inv_scale)
        self.df = prior.df + n_s
        return - self.relative_entropy(prior)

    def relative_entropy(self, othr):
        """Kullback-Leibler divergence between self and othr
        :param othr: single instance of same class as self
        :return: scalar KLdiv(q || p) = E_q{ln q(x) / p(x)}, where q = self and p = othr
        Method: Using notation D = self.n_traits, V = self.prec,
        KLdiv = (q.df - p.df) * E{log_det(self)} / 2
            - (q.df - p.df) * ln(2) * D / 2
            + (p.df * log_det(p.V) - q.df * log_det(q.V)) / 2
            + trace[(inv(p.V) - inv(q.V)) * E{self}] / 2
            + gammaln_D(p.df/2) - gammaln_D(q.df/2)
        where
            E{log_det(self)} = psi_D(q.df/2) + D * ln(2) + log_det(q.V)
            E{self} = q.df * q.V
        = (q.df - p.df) * psi_D(q.df/2) / 2
            + p.df * (log_det(p.V) - log_det(q.V)) / 2
            + trace[(inv(p.V) - inv(q.V)) q.V] * q.df / 2
            + gammaln_D(p.df/2) - gammaln_D(q.df/2)
        """
        d = self.n_traits
        Vpinv_Vq = np.dot(othr.inv_scale, self.scale)
        # ******* try linalg.solve instead   *********
        (s, log_det_VV) = np.linalg.slogdet(Vpinv_Vq)
        return (0.5 * ((self.df - othr.df) * multi_psi(self.df/2, d)
                       - othr.df * s * log_det_VV
                       + (np.trace(Vpinv_Vq) - d) * self.df
                       )
                + multigammaln(othr.df/2, d) - multigammaln(self.df/2, d)
                )

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.scale reduced
        """
        self.scale = self.scale[keep, :][:, keep]

    def standardize(self, s):
        """Standardize for unity trait variance, as estimated externally
        :param s: 1D array with scale factors, one for each trait
        :return: None
        Method: covariance to be divided by s, so precision should be multiplied
        """
        self.scale = self.scale * s * s[:, None]


# -------------------------------------------------------------------------
class TraitPrecisionAmongGroups:
    """Array of independent gamma distributions for model parameter psi,
    representing diagonal prior precision matrix of the group mean trait
    among different ItemResponseGroup instances
    """
    def __init__(self, a=1., b=1.):
        self.a = a
        self.b = np.array(b).reshape((-1,))
        # a is scalar, b must be 1D array

    def __repr__(self):
        return self.__class__.__name__ + f'(a={self.a}, b={self.b})'

    @property
    def prior(self):
        return prior_precision_among(len(self.b))

    @property
    def scale(self):
        return 1. / self.b

    @property
    def inv_scale(self):
        return self.b

    @property
    def mean(self):
        return self.a / self.b

    @property
    def mean_inv(self):
        if self.a > 1:
            return self.b / (self.a - 1)
        else:
            return np.full(self.b.shape, np.nan)

    # @property
    # def var(self):
    #     return self.a / self.b**2

    @property
    def mean_log(self):
        """E{ log( self ) }"""
        return psi(self.a) - np.log(self.b)

    def relative_entropy(self, prior):
        """Kullback-Leibler divergence between GammaRV q and p,
        :param prior: prior with same class and structure
        :return: KLdiv( self || prior ) = E{ ln pdf(x | self) / pdf(x | prior) }_self
        Method:
        pdf(x) = prod_d C(a, b_d) x^{a-1} exp(-b_d x_d)
        C(a, b_d) = b_d^a / Gamma(a)
        """
        pq_b_ratio = prior.b / self.b
        return np.sum(gammaln(prior.a) - gammaln(self.a)
                      - prior.a * np.log(pq_b_ratio)
                      + (self.a - prior.a) * psi(self.a)
                      - self.a * (1 - pq_b_ratio)
                      )

    # ------------------------------------------- MAP learning:
    def adapt(self, mean2_mu):
        """Adapt to given data
        :param mean2_mu: 2D array-like nested list of mean square group property mu
            mean2_mu[g][d] =  E{mu_gd^2} for d-th trait of g-th group
            len(mean2_mu[g]) == n_traits == len(self.b) == number of groups
        :return: ll = KLdiv(self || prior) after adjustment
        2019-08-18, changed for NON-tied self
        """
        prior = self.prior
        mean2 = np.array(mean2_mu)
        (n_groups, n_traits) = mean2.shape
        # self.a = prior.a + n_traits * n_groups / 2
        self.a = prior.a + n_groups / 2
        self.b = prior.b + np.sum(mean2, axis=0) / 2  # sum across g and d
        return - self.relative_entropy(prior)

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.b reduced
        """
        self.b = self.b[keep]

    def standardize(self, s):
        """Standardize for unity trait variance, as estimated externally
        :param s: 1D array with scale factors, one for each trait
        :return: None
        Method: covariance to be divided by mean(s**2)
        """
        self.b = self.b / np.mean(s**2)


# ---------------------------------------- module help functions:

def multi_psi(a, d):
    """Multivariate psi function
    = first derivative of scipy.special.multigammaln(a, d) w.r.t a
    = sum( psi( a + (1-j)/2) for j = 1,...,d ) (Wikipedia)
    = sum( psi( a - (j-1)/2) for j = 1,...,d )
    = sum( psi( a - j/2) for j = 0,...,d-1 )
    """
    return np.sum(psi(a - np.arange(d) / 2))


# ------------------------------------------------------------- Module TEST:

if __name__ == '__main__':
    from scipy.stats import norm

    # --------------------------- simulated trait data:
    n_groups = 2
    n_samples = 10
    n_subjects = 20
    n_traits = 3

    theta = norm.rvs(size=(n_groups, n_samples, n_subjects, n_traits))

    print('*** Testing TraitPrecisionAmongGroups:\n')

    prec_among = prior_precision_among(n_traits)
    print(prec_among)
    print(f'mean= {prec_among.mean}')
    print(f'relative_entropy(prior)=', prec_among.relative_entropy(prec_among.prior))

    mean2_mu = np.mean(theta**2, axis=(1,2))  # mean across samples and subjects
    print('mean2_mu=', mean2_mu)

    ll=prec_among.adapt(mean2_mu=mean2_mu)
    print('adapted: ', prec_among)
    print('ll= ',ll)
    print(f'mean= {prec_among.mean}')
    print(f'relative_entropy(prior)=', prec_among.relative_entropy(prec_among.prior))

    print('\n*** Testing TraitPrecisionWithinGroup:\n')

    prec_within = prior_precision_within(n_traits)
    print(prec_within)
    print(f'mean=\n{prec_within.mean}')
    print(f'mean_inv=\n{prec_within.mean_inv}')
    print(f'mode_inv=\n{prec_within.mode_inv}')
    print(f'relative_entropy(prior)=', prec_within.relative_entropy(prec_within.prior))

    group_cov = np.einsum('gnsi, gnsj -> gsij',
                          theta, theta) / n_samples

    print('adapting to group_cov')
    print('mean empirical group_cov=\n', np.mean(group_cov, axis=(0, 1)))
    print('')

    ll = prec_within.adapt(group_cov)
    print('adapted: ', prec_within)
    print('ll= ', ll)
    print(f'mean=\n{prec_within.mean}')
    print(f'mean_inv=\n{prec_within.mean_inv}')
    print(f'mode_inv=\n{prec_within.mode_inv}')
    print(f'mean_log_det={prec_within.mean_log_det}')
    print(f'relative_entropy(prior)=', prec_within.relative_entropy(prec_within.prior))




""" **** NOT USED ****
Module defining probability distribution of rating-interval thresholds for
response categories in an OrdinalItemResponseModel.

*** Main Class:
OrdinalItemScale --- defining response intervals for one item

Method:
In this version, response thresholds are indirectly determined by vector
w = non-normalized width parameters in the logistic-mapped domain.
The prior distribution of w is gamma,
and the posterior is approximated by a set of equally probable sample vectors.

2019-07-xx testing new version using width w directly as parameter
"""

import numpy as np
# from collections import Counter
# from scipy.optimize import line_search
from scipy.stats import norm
from .safe_logistic import logistic

from samppy import hamiltonian_sampler as ham
from samppy.sample_entropy import entropy_nn_approx as entropy

import logging
logger = logging.getLogger(__name__)

# logger.setLevel(logging.INFO)
logger.setLevel(logging.DEBUG)

ham.VECTOR_AXIS = 1  # just in case default has changed...

# ---------------------------- Module Default Constant:
INITIAL_WIDENING = 1.0  # 1.5
# scale expansion factor for interval bounds,
# after initial settings based on response frequencies,
# to account crudely for fixed logistic scale == 1.

#PRIOR_ETA_SCALE = 1.  # 3.
# scale of Gaussian prior distribution of eta parameters
# defining a diffuse prior

PRIOR_ALPHA = 1.01
# = shape parameter of prior gamma distribution of width parameters
PRIOR_ALPHAm1 = PRIOR_ALPHA - 1.  # only this is needed
PRIOR_BETA = 0.1
# = inverse-scale parameter of prior gamma distribution

PSEUDO_COUNT = 0.5
# = Jeffreys prior for Dirichlet distribution of probability vector,
# given response counts


# --------------------------------------------------------------------
class OrdinalItemScale:
    """Class defining response thresholds,
    and learning of defining parameters,
    for ONE item in an OrdinalItemResponseModel.

    Response interval widths in the logistic-transformed domain
    are proportional to 1D array w with only positive values.
    In each adapt step, the posterior distribution of eta
    is approximated by a set of sample vectors, bounded to be positive
    """

    @classmethod
    def initialize(cls, response_count, n_samples=100):
        """Initialize a scale object crudely to agree with given response counts.
        :param response_count: 1D array with total response counts
            response_count[l] = count of response == l
        :param n_samples: (optional) number of samples
        :return: a single cls object
        """
        f = np.cumsum(response_count + PSEUDO_COUNT)
        p_cum = f / f[-1]
        tau = norm.ppf(p_cum[:-1], scale=INITIAL_WIDENING)
        # tau[l] = upper limit of interval for response==l
        cum_w = np.concatenate(([0.], norm.cdf(tau), [1.]))
        w = np.diff(cum_w)
        # eta = np.log(w)
        # # eta[l] = log width in transformed domain for l-th response interval
        # eta = np.tile(eta, (n_samples, 1))

        return cls(np.tile(w, (n_samples, 1)))

    def __init__(self, w):
        """
        :param w: 2D array with interval widths in logistic-transformed domain
            w[n, l] = n-th sample of width for l-th response interval
        """
        # eta -= np.mean(eta, axis=1, keepdims=True)
        # forced to zero mean
        b = [(0., None) for _ in range(w.shape[-1])]
        # restrict all sample values to be positive
        self.sampler = ham.HamiltonianBoundedSampler(fun=neg_logprob,
                                                     jac=d_neg_logprob,
                                                     x=w,
                                                     bounds=b,
                                                     epsilon=0.2,
                                                     n_leapfrog_steps=10)
        # sampler.args to be defined later
        # NOTE: w is stored only as sampler.x

    def __repr__(self):
        return self.__class__.__name__ + f'(w=\n{repr(self.w)})'

    @property
    def w(self):
        return self.sampler.x

    @property
    def n_samples(self):
        return self.sampler.x.shape[0]

    @property
    def n_response_levels(self):
        return self.sampler.x.shape[1]

    @property
    def tau(self):
        """Response-thresholds samples representing current distribution.
        :return: 2D array tau, with
            tau[m, l] = m-th sample of UPPER threshold for l-th interval,
            NOT including extreme values at -inf and +inf
            tau.shape[-1] == self.n_response_levels - 1
        """
        return _tau(self.w)

    @property
    def typical_trait(self):
        """Crude point estimate of trait parameter, given response,
        estimated at MEDIAN of conditional distribution, given response,
        using current sampled distribution of scale thresholds.
        :return: theta = 2D array, with
            theta[m, l] = m-th sample of typical trait for response == l
            theta.shape[-1] == self.n_response_levels
        """
        cum_w = np.cumsum(self.w, axis=1)
        p = np.concatenate((np.zeros((self.n_samples, 1)), cum_w / cum_w[:, -1:]), axis=1)
        p_mid = (p[:, :-1] + p[:, 1:]) / 2
        return logistic.ppf(p_mid)

    def adapt(self, groups, item_index):
        """One variational-inference update step of sample parameters eta.
        :param groups: dict with elements (group_id, group_model)
            where group_model is a subclass of RespondentGroup object
            that can calculate log-likelihood values for given response thresholds.
        :param item_index: integer index of item with scale defined by self.
        :return: LL = scalar contribution to lower bound to data log-likelihood
        """
        self.sampler.args = (groups, item_index)
        max_steps=20
        try:
            self.sampler.safe_sample(min_steps=5, max_steps=max_steps)
            if self.sampler.n_steps >= max_steps:
                logger.warning(f'Done {self.sampler.n_steps} = MAX allowed sampling steps')
            else:
                logger.debug(f'Done {self.sampler.n_steps} sampling steps')
            if self.sampler.accept_rate > 0.95:  # ****** do this in samppy.safe_sample ********
                self.sampler.epsilon *= 1.3
                logger.debug(f'High sampler accept_rate; increased epsilon = {self.sampler.epsilon}')
        except ham.AcceptanceError:
            logger.warning((f'* AcceptanceError: accept_rate= {self.sampler.accept_rate:.1%} '
                            + f'of {self.sampler.n_trajectories}; '
                            + f'epsilon reduced to {self.sampler.epsilon:.5f}'))
            # ****** keep going anyway ********
        else:
            logger.debug(f'Sampler accept_rate = {self.sampler.accept_rate:.1%}')
            logger.debug(f'Sampler epsilon = {self.sampler.epsilon:.5f}')
        # --------------------------------- DONE distribution of eta
        logger.debug(f'scales[{item_index}] tau percentiles:\n'
                     + np.array2string(np.percentile(self.tau, axis=0, q=[5, 50, 95]),
                                       precision=3))

        # lp = - np.mean(self.sampler.U)
        # the lp part will be included bh RespondentGroup.adapt()
        # so here we contribute only the relative_entropy part
        kl_div = self.relative_entropy_re_prior()
        logger.debug(f'scale[{item_index}]: kl_div={kl_div:.3f}')
        return - kl_div

    def relative_entropy_re_prior(self):
        """KL div(self || prior),
        estimated using current sampled approximation of self.eta
        :return: KL_div = scalar
        Method: E_{self}{ pdf(eta | self) / prior_pdf(eta)}
        """
        return - entropy(self.w) - np.mean(prior_logprob(self.w))


# -------------------------- internal module functions

def neg_logprob(w, groups, item_index):
    """Negative sum of log prob of observed data, given tentative samples.
    :param w: 2D array with tentative sample values
        w[n, l] = n-th sample of l-th width parameter
    :param groups: dict with ref to all response groups
    :param item_index: integer index for this item
    :return: nlp = 1D array
        nlp[n] = negative log-prob for n-th tentative sample
        len(nlp) == w.shape[0]
    """
    # w = np.exp(eta)
    tau = _tau(w)
    lp = sum(g.item_logprob_by_tau(tau, item_index) for g in groups.values())
    if np.any(np.isnan(lp)):
        logger.warning('neg_logprob == nan')
    return - lp - prior_logprob(w)


def d_neg_logprob(w, groups, item_index):
    """Gradient of neg_logprob, given tentative samples.
    :param w: 2D array with tentative sample values
        w[n, l] = n-th sample of l-th width parameter
    :param groups: dict with ref to all response groups
    :param item_index: integer index for this item
    :return: dnlp = 2D array
        dnlp[n, l] = d neg_logprob(w[n]) / d w[n, l]
        dnlp.shape == w.shape
    """
    # w = np.exp(eta)
    tau = _tau(w)
    if np.any(tau == -np.inf) or np.any(tau == np.inf):
        logger.warning('Infinite tau. Should never happen!')
    d_lp_d_tau = sum(g.d_item_logprob_by_tau(tau, item_index) for g in groups.values())
    dlp = np.einsum('mk, mkl -> ml',
                    d_lp_d_tau, _d_tau_dw(w))
    return - dlp - d_prior_logprob(w)


def prior_logprob(w):
    """Prior log pdf of tentative eta array,
    disregarding the normalization constant,
    because the scale parameters are fixed.
    :param w: 2D array of interval-width parameters
        w[n, l] = n-th sample of l-th width
    :return: lp = 1D array,
        lp[n] = log pdf(w[n, :])
        using independent gamma distribution for all elements
    """
    return np.sum(PRIOR_ALPHAm1 * np.log(w) - PRIOR_BETA * w,
                  axis=-1)


def d_prior_logprob(eta):
    """Gradient of prior_logprob for tentative eta array,
    :param eta: 2D array of log interval-width parameters
        eta[n, l] = n-th sample of l-th log-width
    :return: dlp = 2D array,
        dlp[n, l] = d prior_logprob(eta[n,:] / d eta[n, l]
        dlp.shape == eta.shape
    """
    return PRIOR_ALPHAm1 / w - PRIOR_BETA


def _tau(w):
    """Response-interval internal limits, NOT including extremes at -+ inf
    :param w: 2D array of non-normalized interval-width parameters = np.exp(eta)
        w[n, l] = n-th sample of l-th interval non-normalized width in logistic-transformed domain
    :return: t = 2D array with corresponding bounds in the (-inf, inf) range,
        t.shape == (w.shape[0], w.shape[1] - 1)
    """
    # cc = np.cumsum(np.abs(w), axis=1)   # ********** Why abs?
    cc = np.cumsum(w, axis=1)
    return logistic.ppf(cc[:, :-1] / cc[:, -1:])


def _d_tau_dw(w):
    """Jacobian of _tau(w)
    :param w: 2D array of non-normalized interval-width parameters
        w[n, l] = proportional to n-th sample of l-th interval width
        in logistic-transformed domain
    :return: 3D array dt_dw, with
        dt_dw[n, l, j] = d _tau[n, l] / dw[n, j]
        dt_dw.shape == (w.shape[0], w.shape[1]-1, w.shape[1])
    Method:
    t_l = log(w_0 +...+ w_l) - log(w_{l+1} +...+ w_{L-1}); l = 0,..., L-2, any n
    dt_nlj / d w_nj =
        = +1 / (w_0 +...+ w_l); j <= l
        = -1 / (w_{l+1}+...+w_{L-1}); j > l
    2019-07-04 seems OK
    """
    nw = w.shape[1]
    cc = np.cumsum(w, axis=1)
    cc_plus = cc[:, :-1]
    cc_minus = cc[:, -1:] - cc_plus
    mask = np.ones((nw-1, nw))
    dt_dw = (1. / cc_plus[:, :, None] * np.tril(mask, k=0)
             - 1. / cc_minus[:, :, None] * np.triu(mask, k=1))
    return dt_dw


# def _d_tau_d_eta(w):
#     """Jacobian of _tau(w) w.r.t log(w)
#     :param w: 2D array of non-normalized interval-width parameters = np.exp(eta)
#         w[n, l] = proportional to n-th sample of l-th interval width
#         in logistic-transformed domain
#     :return: 3D array dt_d_eta with
#         dt_d_eta[n, l, j] = d _tau[n, l] / d_eta[n, j]
#         dt_d_eta.shape == (w.shape[0], w.shape[1]-1, w.shape[1])
#     2019-07-04, sum(_d_tau_d_eta(w), axis=-1) == 0. as it should
#     """
#     return _d_tau_dw(w) * w[:, None, :]
#

# --------------------------------------------------- TEST:

if __name__ == '__main__':
    # from scipy.optimize import check_grad, approx_fprime
    from scipy.stats import gamma

    g = gamma(a=PRIOR_ALPHA, scale=1. / PRIOR_BETA)
    print(f'prior width range: {g.ppf(0.01):.3f} ... {g.ppf(0.99):.3f}')
    prior_width_ratio = g.ppf(0.99) / g.ppf(0.01)
    print(f'prior gamma width ratio = {prior_width_ratio:.2f}')

    w = np.array([[1., 2., 3., 4., 5.]])
    w = np.concatenate((w, 2*w), axis=0)
    # 2 samples, 5 levels
    print('_d_tau_dw(w)= ', _d_tau_dw(w))

    # dt_d_eta = _d_tau_d_eta(w)
    # print('_d_tau_d_eta(w)= ', dt_d_eta)
    # print('sum dt_d_eta= ', np.sum(dt_d_eta, axis=-1))

    item_sc = OrdinalItemScale.initialize(10 * w[0], n_samples=3)
    print('item_sc= ', item_sc)
    print('item_sc.n_samples=', item_sc.n_samples)
    print('item_sc.n_response_levels=', item_sc.n_response_levels)
    print('item_sc.w=\n', item_sc.w)
    print('item_sc.tau=\n', item_sc.tau)
    print('item_sc.typical_trait=\n', item_sc.typical_trait)

    # ------- neg_loprob gradient tested in test_item_scale



"""Numerically safer version of scipy.stats.logistic distribution.
May be replaced by future improved version of scipy.stats.logistic.

Arne Leijon, 2017-03-13
2019-04-28, fix method signature to agree with scipy.stats.rv_continous, no real change
"""
import numpy as np
from scipy.stats import rv_continuous
from scipy.special import expit, logit


class logistic_gen(rv_continuous):
    """Safer version with central methods re-implemented.
    Method signatures copied from scipy.stats.logistic_gen
    """
    def _logpdf(self, x, *args):
        """logistic pdf(x) = exp(-x)/(1+exp(-x))**2
        Thus, logpdf(x) = -x - 2 * log((1+exp(-x))
        This is fine for large positive x, but pdf is symmetric,
        so we use abs(x) to avoid overflow for large negative x.
        :param x: array-like input values
        :param args: not used
        :return: lp = array of logpdf values
            lp.shape == x.shape
        """
        nax = -np.abs(x)
        return nax - 2. * np.log1p(np.exp(nax))

    def _logcdf(self, x, *args):
        """log logistic.cdf = -log(1 + exp(-x))
        = -x - log(1 + exp(x)), for x < 0.
        safer than scipy.stats.logistic.logcdf
        :param x: array-like input values
        :param args: not used
        :return: lp = array of logcdf values
            lp.shape == x.shape
        """
        x = np.asarray(x)
        ax = np.abs(x)
        lP = -np.log1p(np.exp(-ax))
        negx = x < 0
        lP[negx] -= ax[negx]
        return lP

    # ----------------------------------------------------
    # following private methods directly copied from scipy.stats.logistic_gen

    def _rvs(self):
        return self._random_state.logistic(size=self._size)

    def _pdf(self, x, *args):
        return np.exp(self._logpdf(x))

    def _cdf(self, x, *args):
        return expit(x)

    def _ppf(self, q, *args):
        return logit(q)

    def _sf(self, x, *args):
        return expit(-x)

    def _isf(self, q, *args):
        return - logit(q)

    def _stats(self):
        return 0, np.pi * np.pi / 3.0, 0, 6.0 / 5.0

    def _entropy(self):
        # http://en.wikipedia.org/wiki/Logistic_distribution
        return 2.0
# ---------------------------------------------------------


logistic = logistic_gen(name='logistic')


# --------------------------------------------------------- TEST:
if __name__ == '__main__':
    from scipy.stats import logistic as sc_logistic

    x = np.array([0., 2., 1000., np.inf])

    print('x= ', x)
    print('\nImproved logistic version:')
    print('logistic.logpdf(x) = ', logistic.logpdf(x))
    print('logistic.logpdf(-x)= ', logistic.logpdf(-x))
    print('logistic.pdf(x) = ', logistic.pdf(x))
    print('logistic.pdf(-x)= ', logistic.pdf(-x))
    print('logistic.cdf(x) = ', logistic.cdf(x))
    print('logistic.cdf(-x)= ', logistic.cdf(-x))
    print('logistic.logcdf(x) = ', logistic.logcdf(x))
    print('logistic.logcdf(-x)= ', logistic.logcdf(-x))
    print('logistic.ppf(cdf(x)) = ', logistic.ppf(logistic.cdf(x)))
    print('logistic.ppf(cdf(-x))= ', logistic.ppf(logistic.cdf(-x)))

    print('logistic.pdf(x, loc=2.)', logistic.pdf(x, loc=2.))
    print('logistic.logcdf(x, scale=0.02)', logistic.logcdf(x, scale=0.02))

    print('logistic.rvs(size=10)', logistic.rvs(size=10))

    print('')
    print('scipy.stats version:')
    print('scipy.stats.logistic.logpdf(x) = ', sc_logistic.logpdf(x))
    print('scipy.stats.logistic.logpdf(-x)= ', sc_logistic.logpdf(-x))
    print('scipy.stats.logistic.logcdf(x) = ', sc_logistic.logcdf(x))
    print('scipy.stats.logistic.logcdf(-x)= ', sc_logistic.logcdf(-x))





"""Interface to read Item Response Data stored in an SPSS (.sav) file.
NOT RECOMMENDED, better to use SPSS to export file in another format

*** MAIN CLASS:
Table --- iterable interface to an SPSS file,
    accessing fields by their names as used in the database.

*** NOTE: MacOS savReaderWriter needs user-assigned Region and Preferred Language
in System Preferences to match an existing locale setting.
Otherwise,
locale.setlocale(locale.LC_ALL, '') will raise an exception
when savReaderWriter tries to open and/or close an SPSS file.
"""
# *********************** check savReader.close locale again ******************

import savReaderWriter
# ***** NOTE symbolic links in /usr/local/lib are NEEDED,
# not correctly installed by PIP package installer

from . import table_reader


# -----------------------------------------------------------------
class Table(table_reader.Table):
    """Reader for tabular data stored in SPSS file,
    using package savReaderWriter
    """
    # superclass __init__ is sufficient, no new parameters

    # def get_header(self):
    #     """Connect to the file and read file header
    #     :return: list with header name strings from the file
    #     """
    #     with self.open_reader(return_header=True) as spss_reader:
    #         return spss_reader.next()
    #
    # def open_reader(self, return_header=False):
    #     """Open file for read-only access.
    #     :param return_header: boolean
    #     :return: iterator object over file rows,
    #         typically used in a with statement.
    #         The reader.next() method yields each row as stored in the file
    #         without any type cast or other checks
    #     """
    #     return savReaderWriter.SavReader(self.file_path,
    #                                      returnHeader=return_header,
    #                                      selectVars=self.fields,
    #                                      ioUtf8=True)

    def __iter__(self):
        """Generator of rows for read-only access.
        :return: iterator object over file rows, except header
        """
        with savReaderWriter.SavReader(self.table_id,
                                       returnHeader=False,
                                       selectVars=self.fields,
                                       ioUtf8=True) as reader:
            for (row_n, row) in enumerate(reader):
                if row_n % self.sample_factor == 0:
                    row = self.process_row(row)
                    if row is not None:
                        yield row

    def preview(self, max_records=10):
        """Generator of rows for read-only access.
        :return: iterator object over file rows, except header
        """
        # result = []
        with savReaderWriter.SavReader(self.table_id,
                                       returnHeader=True,
                                       # selectVars=self.fields,
                                       ioUtf8=True) as reader:
            for (row_n, row) in enumerate(reader):
                if row_n > max_records:
                    break
                yield row
        # return result


# -----------------------------------------------------------------------------
# if __name__ == '__main__':
#     from pathlib import Path
#
#     HAQ_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry'
#     # top-dir for everything to be used here
#
#     ioiha_data_path = HAQ_PATH / 'IRT-IOI-HA' / 'IOI-HA-data'
#     # to ioi-ha data sets OTHER THAN the Swedish Quality Registry
#
#     ioiha_kramer = ioiha_data_path / 'Kramer'
#
#     kramer_file = ioiha_kramer / 'moeder_final - kopieforArne.sav'
#     kramer_file = ioiha_kramer / 'NLSH-Kramer-T2IOI-HA.sav'
#
#     # help(savReaderWriter.SavReader)
#
#     with savReaderWriter.SavReader(kramer_file, returnHeader=True,
#                                    # rawMode=True,
#                                    # selectVars=['ioi_v1', 'ioi_v2'],
#                                    ioUtf8=True) as reader:
#         header = reader.next()
#         print('header= ', header)
#         for (n, line) in enumerate(reader):
#             print(line)
#             if n > 10:
#                 break
#
#     # spss = Table(kramer_file)
#     # print('spss.header = ', spss.header)
#     #
#     # data = spss.preview(10)
#     # print('spss.preview(10) =\n', data)
#
#     # -------------------------------- test with missing-values recoding:
#     spss = Table(kramer_file,
#                  # fields=['ioi_v1', 'ioi_v2', 'ioi_v3'],
#                  fields=[f'T2IOIHA{i+1}' for i in range(7)],
#                  missing_values=[9999999.0],
#                  field_types=str)
#     print('repr(spss)= ', spss)
#
#     for (n, row) in enumerate(spss, start=1):
#         if n > 10:
#             break
#         print(f'row #{n} = {row}')
#
#     print('\n *** spss.preview(10)=\n', spss.preview(10))
#
#     # -------------------------------------------------------------------
#     print('\n*** final test with missing_values, type_cast, recode_fcn ***\n')
#
#     def recode_missing(row):
#         for i in range(len(row)):
#             if row[i] is None:
#                 row[i] = 0
#
#     def recode_fields14(row):
#         """IOI-HA items 1,2,3,4 coded as 0-4,
#         recode to 1-5,
#         BEFORE recoding missing-data as zero.
#         """
#         for i in [1, 4]:
#             if row[i-1] is not None:
#                 row[i-1] += 1
#     # --------------------------------------------
#
#     spss = Table(kramer_file,
#                  # fields=['ioi_v1', 'ioi_v2', 'ioi_v3'],
#                  fields=[f'T2IOIHA{i+1}' for i in range(7)],
#                  missing_values=[9999999.0],
#                  field_types=int,
#                  recode_fcns=[recode_fields14, recode_missing])
#     for (n, row) in enumerate(spss, start=1):
#         # if any(r_i is not None and (r_i <= 0 or r_i >= 5) for r_i in row):
#         if any(r_i is not None and (r_i > 5) for r_i in row):
#             print(f'row #{n} = {row}')
#
#     print(f'len(spss)={len(spss)}')
#     n_all_zero = sum(all(r_i == 0 for r_i in r)
#                      for r in spss)
#     print(f'Number of all-zero records = {n_all_zero}')
#



"""Classes to handle mapping between questionnaire items and subjective traits

*** Main Classes:
TraitSelector: Probability distribution of a one-of-D boolean switch array
TraitProb: Distribution of prior probability mass vector for any TraitSelector object
"""
import numpy as np
from scipy.special import gammaln, psi

# ----------------------------------------------------------
PRIOR_TRAIT_CONCENTRATION = 0.001
# = prior concentration parameters for TraitProb Dirichlet distribution


# ----------------------------------------------------------
class TraitSelector:
    """Probability distribution of a one-of-D boolean switch array
    z = (z_1, ..., z_D) with only ONE element z_d == True,
    used as property trait_sel of ONE OrdinalItemScale object
    indicating that responses to this object
    are determined by the the d-th trait.
    """
    def __init__(self, weight):
        """
        :param weight: 1D array with un-normalized probability mass elements
        """
        weight = np.array(weight)
        self.prob = weight / np.sum(weight)

    @property
    def mean(self):  # ******* mean_responsibility!
        return self.prob

    def adapt(self, log_responsibility, prior):
        """Adapt self.prob for given trait responsibility
        :param log_responsibility: 1D array with sum logprob(responses) given each trait
        :param prior: single TraitProb object
        :return: None
        """
        log_r = log_responsibility + prior.mean_log
        log_r -= np.amax(log_r)
        r = np.exp(log_r)
        self.prob = r / np.sum(r)

    def relative_entropy_re_prior(self, prior):
        """KL div = E{ log q(z) / p(z | w) }_{z, w}
        :param prior: single TraitProb object
        :return: scalar KL-div
        Method:
        q(z) = prod_t r_t^{z_t}; p(z | w) = prod_t w_t^{z_t}
        where r = self.prob; w = prior
        """
        eps = np.finfo(float).tiny  # just to prevent log(0.)
        return np.dot(self.prob, np.log(self.prob + eps) - prior.mean_log)

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.prob reduced, and still normalized
        """
        w = self.prob[keep]
        self.prob = w / np.sum(w)


# ---------------------------------------------------------------------------
class TraitProb:
    """Distribution of hierarchical prior probability mass for all TraitSelector objects,
    modeled as a single Dirichlet distribution.
    """
    @classmethod
    def initialize(cls, n_traits):
        """Create prior trait distribution
        :param n_traits: scalar integer number of traits
        :return: single initialized TraitProb object
        """
        alpha = PRIOR_TRAIT_CONCENTRATION * np.ones(n_traits)
        # ****** decreasing alpha by trait index ???
        return cls(alpha)

    def __init__(self, alpha):
        self.alpha=np.array(alpha)

    def __repr__(self):
        return (self.__class__.__name__
                + '(' + f'alpha= ' + repr(self.alpha) + ')')

    @property
    def mean(self):
        return self.alpha / np.sum(self.alpha)

    @property
    def mean_log(self):
        """E{ log W } where W is random vector represented by self"""
        return psi(self.alpha) - psi(np.sum(self.alpha))

    def adapt(self, sum_resp):
        """adapt from given trait-responsibility sum
        :param sum_resp: sum of trait responsibilities across items
        :return: - kl_div
        """
        self.alpha = sum_resp + self.initialize(len(self.alpha)).alpha
        return - self.relative_entropy_re_prior()

    def relative_entropy_re_prior(self):
        """KL_div = E{ log q(w) / p(w) }_{q(w}), where
        q(w) = self distribution; p(w) = self.initialize
        :return: scalar KL_div
        Method:
        q(w) = C(q_alpha) prod_t w_t^{q_alpha - 1}
        p(w) = C(p_alpha) prod_t w_t^{p_alpha - 1}
        C(alpha) = log(gamma(sum alpha)) - sum log(gamma(alpha))
        """
        p = self.initialize(len(self.alpha))
        # = prior
        kl_div = (np.dot(self.alpha - p.alpha, self.mean_log)
                  + gammaln(np.sum(self.alpha)) - gammaln(np.sum(p.alpha))
                  + np.sum(gammaln(p.alpha) - gammaln(self.alpha)))
        return kl_div

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.trait_sel pruned
        """
        self.alpha = self.alpha[keep]




"""Module defining a Bayesian IRT model for the responses of a group of subjects
to a questionnaire with several items.
Each item response is assumed to be determined by the outcome of a
latent random variable with a subject-specific location.

Each uni-dimensional subject trait may determine responses to one or several items.

This model is implemented only for data with ORDINAL responses.

*** Main Classes:

ResponseGroup -- model for a group of subjects for which
    response data are available for each item for each subject.

GroupMeanTrait -- model for Gaussian distribution of mean trait vector for one group

Arne Leijon, 2019-07-08, first version
2019-07-23, cannot use SummaryResponseGroup!
2019-08-11, delete class SummaryResponseGroup
2019-08-11, use scales.trait_sel.mean as item_trait_weight
2019-08-15, new initialize method, incl. theta randomization
2019-08-17, single ResponseGroup class, generalized for tied or non-tied prec_among
2019-08-20, ResponseGroup.adapt returns self, to allow multi-processing
"""
import numpy as np
# from scipy.stats import multivariate_normal  # replaced by np.random.Generator

from samppy import hamiltonian_sampler as ham
from samppy.sample_entropy import entropy_nn_approx as entropy
from .ir_predictive import GaussianGivenParam

from .safe_logistic import logistic

import logging
logger = logging.getLogger(__name__)

logger.setLevel(logging.INFO)
# logger.setLevel(logging.DEBUG)

ham.VECTOR_AXIS = 1  # just in case default has changed...

multivariate_normal = np.random.default_rng().multivariate_normal
# = random generator for multivariate Gaussian vector

# ----------------------------------------------------------


# ---------------------------------------------- Respondent Group Model:

class ResponseGroup:
    """Model representing subjects' latent traits in an OrdinalItemResponseModel,
    learned by adaptation to individual item responses for each subject.
    The trait distribution is represented by sampling.
    """
    @classmethod
    def initialize_by_item(cls, subject_responses, scales, trait_scale,
                           name,
                           n_samples=20):
        """Initialize trait values by sampling, on trait for each item
        :param subject_responses: iterable yielding subject responses as a list with
            g_subjects[s][i] = integer index of response to i-th item,
            coded with origin==1 for the first response alternative,
            and missing response encoded as 0.
            Recoded later to python origin 0 for internal use.
        :param scales: list of ItemScale instances, one for each item
        :param trait_scale: scalar initially assumed scale of trait values
        :param n_samples: number of samples of the intra-subject distribution of trait values
        :param name: string identifying this group
        :return: single cls instance
        """
        # collect all response data in one array
        s_responses = np.array([s_r for s_r in subject_responses]).T
        # s_responses[i, s] = response-level index for i-th item by s-th subject
        # still with origin == 1, missing response == 0
        # n_items = len(s_responses)
        th = np.array([sc_i.sample_trait(r_i - 1, n_samples, trait_scale=trait_scale)
                       for (sc_i, r_i) in zip(scales, s_responses)])
        # th[i, m, s] = m-th sample for i-th item and s-th subject
        th = th.transpose((1, 2, 0))
        # th[m, s, i] = m-th sample of i-th item for s-th subject
        mu = GroupMeanTrait(loc=np.mean(th, axis=(0, 1)))
        logger.info(f'Initializing {repr(name)}'
                    + f' with {s_responses.shape[-1]} subjects')
        logger.debug('Initial item mu.loc= ' + np.array2string(mu.loc, precision=3))
        return cls(subject_responses=s_responses,
                   theta=th,
                   mu=mu,
                   name=name)

    def __init__(self, subject_responses, theta, mu, name):
        """
        :param subject_responses: 2D array-like list with response data
            response[i, s] = integer ordinal response index for i-th item by s-th subject
            coded with origin==1 for the first response alternative,
        :param theta: 3D array with initial subject latent trait samples
            theta[n, s, t] = n-th sample of t-th trait for s-th subject
        :param mu: GroupMeanTrait instance representing current distribution of mean
        :param name: string with arbitrary label for this object
        NOTE: Internal property
        _response = 2D array of individual response indices, with
            response[i, s] = integer index of response to i-th item,
            encoded with origin==0, and missing response encoded as -1.
        The corresponding external property is
        subject_responses = _response + 1,
        with origin == 1 and missing response == 0, for external presentation.
        """
        assert theta.ndim == 3, 'theta must be 3-dim numpy array'
        self._response = np.array(subject_responses) - 1
        (n_samples, self.n_subjects, self.n_traits) = theta.shape
        self.sampler = ham.HamiltonianSampler(fun=self.neg_logprob,
                                              jac=self.d_neg_logprob,
                                              x=theta.reshape(n_samples, -1),
                                              epsilon=0.3,
                                              n_leapfrog_steps=10)
        # NOTE: theta is stored only as sampler.x
        # args for neg_logprob and d_neg_logprob are defined later, before each call
        self.mu = mu
        self.name = name
        self.ll = None  # log likelihood saved here by adapt step

    def __repr__(self):
        return (self.__class__.__name__ + f'(name= {repr(self.name)},'
                + f'\n\t subject_responses= array with {len(self._response)} responses'
                + f' by {self.n_subjects} subjects,'
                + f'\n\t theta= array with shape= {self.theta.shape},'
                + '\n\t mu= ' + self.mu.__repr__()
                + ')')

    def cov_all(self):
        """Covariance matrix across all samples, all subjects,
        used only for initial rotation of trait dimensions
        as determined externally
        :return: c = 2D array covariance matrix
        """
        # m = np.mean(self.theta, axis=(0,1))
        logger.debug('Initial cov without mean subtraction')
        x = self.theta.reshape((-1, self.theta.shape[-1]))  # ***** - m
        c = np.cov(x, rowvar=False)
        # (e_val, e_vec) = np.linalg.eigh(c)  # **** only for test
        return c

    def transform_traits(self, proj):
        """Transform initial item-traits to (sub-)space defined by external PCA
        :param proj: 2D projection matrix with orthonormal COLUMN vectors
            proj[:, t] = unit vector for t-th trait
        :return: None
        self.theta and self.mu transformed to proj coordinate system
        """
        th = np.dot(self.theta, proj)
        self.n_traits = th.shape[-1]
        self.theta = th
        self.mu = GroupMeanTrait(loc=np.mean(th, axis=(0, 1)))
        logger.info(f'Transformed group {repr(self.name)} self.mu.loc= '
                    + np.array2string(self.mu.loc, precision=3, suppress_small=True))

    @property
    def subject_responses(self):
        """Subject responses as obtained from input files
        using the external encoding {1,..., L_i},
        with missing response == 0
        :return: r = 2D array with responses,
            r[s, i] = response by s-th subject to i-th item
        """
        return self._response + 1

    @property
    def n_samples(self):
        return self.sampler.x.shape[0]

    @property
    def theta(self):
        """Sample representation of current distribution of trait values
        for all subjects in this group
        :return: theta = 3D array of samples
            theta[n, s, t] = n-th sample of t-th trait for s-th subject
        """
        return self.sampler.x.reshape((self.n_samples, self.n_subjects, self.n_traits))

    @theta.setter
    def theta(self, th):
        """Store theta values in 2D array sampler.x
        :param th: 3D theta matrix
            th[n, s, t] = n-th sample of t-th trait for s-th subject
        :return: None
        """
        self.sampler.x = th.reshape((self.n_samples, -1))

    @property
    def mean_cov_theta(self):
        """Mean cov of current theta distribution for each subject
        averaged across samples.
        :return: c = 3D array with
        c[s] = mean covariance matrix for s-th subject =
        = E{ theta_s - mu)^T (theta_s - mu) }
        = E{ theta_s - mu.loc)^T (theta_s - mu.loc) } + mu.cov
        """
        th_diff = self.theta - self.mu.loc
        c = np.einsum('nsi, nsj -> sij',
                      th_diff, th_diff) / th_diff.shape[0]
        return c + self.mu.cov

    @property
    def entropy_theta(self):
        """Entropy of current theta distribution,
        estimated from samples by nearest-neighbor method
        :return: h_sum = scalar entropy, sum across all subjects
        """
        th = self.theta
        # h = entropy(self.sampler.x)
        # h = single estimate using all subjects in one operation.
        # But we know distributions are independent among subjects,
        # and Singh et al. (2016) showed convergence is slower with increasing dimension,
        # so an average across all subjects, at low dimension for each subject,
        # is most probably closer to the true value.
        h_sum = sum(entropy(th[:, s, :]) for s in range(th.shape[1]))
        return h_sum

    def adapt(self, scales, precision_within, precision_among):
        """One learning step for distribution of group theta values,
        and group mu property,
        using responses data and current distribution of item scale parameters.
        :param scales: list of ItemScale objects, one for each item
        :param precision_within: TraitPrecisionWithinGroup object defining
            current distribution of inter-individual precision matrix within every group
        :param precision_among: TraitPrecisionAmongGroups object defining
            current distribution of psi = scalar precision among group means
        :return: self updated, to allow multi-processing
        """
        _id = repr(self.name)  # used for logger output
        logger.debug(self.__class__.__name__ + f'[{_id}].adapt(...)')
        # ---------------------------------- adapt self.theta by Hamilton sampling
        tau = [sc.tau for sc in scales]
        weight = [sc.trait_sel.mean for sc in scales]
        # response thresholds, fixed during theta learning
        if np.any([np.any(np.isinf(tau_i)) for tau_i in tau]):
            logger.warning(f'{_id}: Infinite tau. Should not happen!')
        mean_precision_within = precision_within.mean
        mean_log_det_prec_within = precision_within.mean_log_det
        # = external parameters, constant during re-sampling of theta
        # but change between calls
        self.sampler.args = (tau, weight, mean_precision_within, mean_log_det_prec_within)
        max_steps=20
        # logger.debug(f'test sampler._rng.std_normal: {self.sampler._rng.standard_normal()}')
        try:
            self.sampler.safe_sample(min_steps=5, max_steps=max_steps)
            if self.sampler.n_steps >= max_steps:
                logger.warning(f'{_id}: Done {self.sampler.n_steps} = MAX allowed sampling steps')
            else:
                logger.debug(f'{_id}: Done {self.sampler.n_steps} sampling steps')
        #     if self.sampler.accept_rate > 0.95:  # ****** do this in samppy.safe_sample ??? ********
        #         self.sampler.epsilon *= 1.3
        #         logger.debug(f'{_id}: High sampler accept_rate; increased epsilon = {self.sampler.epsilon}')
        except ham.AcceptanceError:  # raised by sampler, even after reducing epsilon
            # logger.debug(f'* AcceptanceError {self}.adapt()')
            # self.sampler.epsilon *= (0.7 + 0.2 * uniform.rvs())  # ************** SKIP
            logger.warning((f'{_id}: AcceptanceError: accept_rate= {self.sampler.accept_rate:.2f} ' +
                            f'of {self.sampler.n_trajectories}; ' +
                            f'epsilon reduced to {self.sampler.epsilon:.5f}'))
            # ****** keep going anyway ********
        else:
            logger.debug(f'{_id}: Sampler accept_rate = {self.sampler.accept_rate:.1%}')
            logger.debug(f'{_id}: Sampler epsilon = {self.sampler.epsilon}')
        # --------------------------------- DONE distribution of theta
        LL = - np.mean(self.sampler.U)
        LL += self.mu.adapt(np.mean(self.theta, axis=0),
                            mean_precision_within,
                            precision_among)
        h = self.entropy_theta
        logger.debug(f'{_id}: LL={LL:.1f}, + entropy={h:.1f}')
        self.ll = LL + h
        logger.debug(self.__class__.__name__ + f'[{_id}].adapt(...) finished')
        return self

    def neg_logprob(self, x, tau, weight, mean_prec, mean_log_det_prec):
        """Negative log-likelihood for any sample subject trait vector
        :param x: 2D array of trait values as stored in self.sampler.x
            x[n, st] = n-th sample of (s,t)-th subject,trait value
        :param tau: list of 2D arrays with UPPER item response thresholds
            tau[i][m, l] = m-th sample of UPPER threshold for l-th response interval of i-th item
            EXCEPT the extreme at +inf
            tau[i].shape[-1] == L_i - 1; L_i = n ordinal response levels
        :param weight: list of 1D arrays of item trait weights
            weight[i][t] = i-th item weight for t-th trait
        :param mean_prec: 2D symmetric intra-group mean precision matrix
        :param mean_log_det_prec: scalar log det(precision matrix)
        :return: nlp = 1D vector of negative log-likelihood values
            nlp.shape == (x.shape[0],) == (self.n_samples,)
        """
        th = x.reshape((self.n_samples, self.n_subjects, self.n_traits))
        return (- self.logprob_by_theta(th, tau, weight)
                - self.prior_logprob(th, mean_prec, mean_log_det_prec))

    def d_neg_logprob(self, x, tau, weight, mean_prec, mean_log_det_prec):
        """Gradient of logprob for any sample subject trait vector
        :param x: 2D array of trait values as stored in self.sampler.x
            x[n, st] = n-th sample of (s,t)-th subject,trait value
        :param tau: list of 2D arrays with UPPER item response thresholds
            tau[i][m, l] = m-th sample of UPPER threshold for l-th response interval of i-th item
            EXCEPT the extreme at +inf
            tau[i].shape[-1] == L_i - 1; L_i = n ordinal response levels
        :param weight: list of 1D arrays of item trait weights
            weight[i][t] = i-th item weight for t-th trait
        :param mean_prec: 2D symmetric intra-group mean precision matrix
        :param mean_log_det_prec: scalar log det(precision matrix)
            NOT NEEDED to calculate gradient w.r.t. theta
        :return: dlp = 2D array of gradient row vectors, one for each sample
            dlp.shape == x.shape
        """
        th = x.reshape((self.n_samples, self.n_subjects, self.n_traits))
        dnlp = (- self.d_logprob_by_theta(th, tau, weight)
                - self.d_prior_logprob(th, mean_prec))
        return dnlp.reshape((self.n_samples, -1))

    def prior_logprob(self, theta, mean_prec, mean_log_det_prec):
        """Prior log-probability of theta, given current group mu
        :param theta: 3D reshaped array of traits
            theta[n, s, t] = n-th sample of t-th trait for s-th subject
        :param mean_prec: 2D array = mean precision matrix within this group
            externally calculated
        :param mean_log_det_prec: 2D mean log det precision matrix
            externally calculated
        :return: lp = 1D array, with lp[n] = n-th sample
        Method: prior theta_s is Gaussian with location self.loc and external prec matrix
        for each subject.
        """
        # th_diff = theta.reshape(self.n_samples, self.n_subjects, -1) - self.mu.loc  # ********
        th_diff = theta - self.mu.loc
        return 0.5 *(self.n_subjects * mean_log_det_prec
                     - np.einsum('nsi, ij, nsj -> n',
                                 th_diff, mean_prec, th_diff))

    def d_prior_logprob(self, theta, mean_prec):
        """Gradient of prior_log-probability of theta, given current group mu
        :param theta: 3D reshaped array of traits
            theta[n, s, t] = n-th sample of t-th trait for s-th subject
        :param mean_prec: 2D mean precision matrix within this group
        :return: dlp = 2D array with gradient sample vectors
            dlp.shape == theta.shape
        """
        # th_diff = theta.reshape(self.n_samples, self.n_subjects, -1) - self.mu.loc   # ******** again
        th_diff = theta - self.mu.loc
        # th_diff[n, s, t] = n-th sample, s-th subject, t-th trait
        # dlp_old = - np.einsum('nsi, ij -> nsj',
        #                       th_diff, mean_prec).reshape(theta.shape)
        dlp = - np.dot(th_diff, mean_prec)  # .reshape(theta.shape)
        return dlp

    def logprob_by_theta(self, theta, tau, weight):
        """log-likelihood of observed responses for any sample subject trait vector
        summed across all responses, averaged across threshold samples
        :param theta: 3D reshaped array of traits
            theta[n, s, t] = n-th sample of t-th trait for s-th subject
        :param tau: list of 2D arrays of response thresholds, EXCEPT extremes
            tau[i][m, l] =  m-th sample of UPPER interval limit for l-th response to i-th item
            tau[i].shape[-1] == L_i - 1; L_i = number of ordinal response levels
        :param weight: list of 1D arrays of item trait weights
            weight[i][t] = i-th item weight for t-th trait
        :return: lp = 1D vector of log-likelihood values
            lp.shape == (self.n_samples,)
        """
        # th = theta.reshape((self.n_samples, self.n_subjects, self.n_traits))  # **** already done !
        lp = [item_logprob_by_theta(theta, tau_i, w_i, r_i)
              for (tau_i, w_i, r_i) in zip(tau,
                                           weight,
                                           self._response)]
        # lp[i, n] = logprob for n-th theta sample for i-th item
        # lp = np.dot(self.trait_item_map, lp)
        return np.sum(lp, axis=0)

    def d_logprob_by_theta(self, theta, tau, weight):
        """Gradient of logprob w.r.t theta, for any given sample subject trait vector
        :param theta: 3D reshaped array of traits
            theta[n, s, t] = n-th sample of t-th trait for s-th subject
        :param tau: list of 2D arrays of response thresholds, EXCEPT extremes
            tau[i][m, l] =  m-th sample of UPPER interval limit for l-th response to i-th item
            tau[i].shape[-1] == L_i - 1; L_i = number of ordinal response levels
        :param weight: list of 1D arrays of item trait weights
            weight[i][t] = i-th item weight for t-th trait
        :return: dlp = 3D array of gradients, one for each sample
            dlp[n, s, t] = n-th sample gradient for t-th trait of s-th subject
            dlp[n, s, t] = d logprob(theta[n]) / d theta[n, s, t]
            dlp.shape == theta.shape
        """
        # th = theta.reshape((self.n_samples, self.n_subjects, self.n_traits))  # **** already done !
        dlp = [d_item_logprob_by_theta(theta, tau_i, w_i, r_i)
               for (tau_i, w_i, r_i) in zip(tau,
                                            weight,
                                            self._response)]
        # dlp[i, n, s, t] = n-th sample of grad logprob w.r.t theta for i-th item, s-th subject
        # dlp = np.einsum('ti, ins -> nst',
        #                 self.trait_item_map, dlp)
        dlp = np.sum(dlp, axis=0)
        # dlp[n, s, t] = n-th d_logprob_by_theta sample for t-th trait, s-th subject
        return dlp  # .reshape(dlp.shape[0], -1)  # shape == theta.shape

    def item_logprob_by_tau(self, tau_i, i, w_i=None):
        """log prob of observed responses for ONE item,
        summed across subjects, averaged across theta samples
        using current theta distribution.
        :param tau_i: tau_i[m, l] = m-th sample of UPPER interval limit for l-th response
            tau_i.shape[-1] == L_i - 1
        :param i: index of selected item
        :param w_i: (optional) 1D array with trait weights for this item
        :return: lp = 2D array, with lp[m, t] = m-th sample for t-th trait
        """
        r_i = self._response[i]
        # theta_i = self.theta[..., self.item_trait_index[i]]
        # theta_i[n, s] = n-th trait sample for s-th subject for this item
        (tau_low, tau_high) = response_interval(tau_i, r_i)
        # tau_low[m, s] and tau_high[m, s] are (low, high) interval limits
        th = self.theta[:, None, :, :]  # to allow broadcast
        # th[n, 0, s, t] = n-th sample of t-trait for s-th subject
        lp = log_prob_range(tau_low[..., None] - th, tau_high[..., None] - th)
        # lp[n, m, s, t] = n-th theta sample, m-th tau sample for s-th subject
        lp = np.sum(lp, axis=-2)
        lp = np.mean(lp, axis=0)
        # lp[m, t] = m-th tau sample, for t-th trait
        if w_i is None:
            return lp
        else:
            return np.dot(lp, w_i)

    def d_item_logprob_by_tau(self, tau_i, i, w_i):
        """Gradient of item_logprob_by_tau w.r.t. tau_i,
        summed across subjects, averaged across theta samples
        :param tau_i: tau_i[m, l] = m-th sample of UPPER interval limit for l-th response
            tau_i.shape[-1] == L_i - 1
        :param i: index of selected item
        :param w_i: (optional) 1D array with trait weights for this item
        :return: dlp = 3D array, with
            dlp[m, l, t] = d logprob(tau_i[m, :] / d tau_i[m, l], given t-th trait
            dlp.shape == (*tau_i.shape, self.n_traits)
        """
        r_i = self._response[i]
        # theta_i = self.theta[..., self.item_trait_index[i]]
        (tau_low, tau_high) = response_interval(tau_i, r_i)
        # tau_low[m, s] and tau_high[m, s] are (low, high) interval limits
        th = self.theta[:, None, :, :]  # to allow broadcast
        (da, db) = d_log_prob_range_dab(tau_low[..., None] - th, tau_high[..., None] - th)
        # da[n, m, s, t] = n-th theta sample, m-th tau sample for t-th trait of s-th subject
        # of derivative w.r.t. tau_low
        # db similar for tau_high
        da = np.dot(np.mean(da, axis=0), w_i)
        db = np.dot(np.mean(db, axis=0), w_i)
        # da[n, m, s] = n-th theta sample, m-th tau sample of s-th subject, same for db
        n_response_levels = tau_i.shape[-1] + 1
        r_bool = np.array([r_i == l
                           for l in range(n_response_levels)]).T
        dlp = np.dot(da, r_bool[:, 1:]) + np.dot(db, r_bool[:, :-1])
        # dlp[m, l] = derivative w.r.t tau_i[m, l], sum across all responses
        # when tau[m, l] is LOWER limit PLUS when it is UPPER limit
        return dlp

    # -------------------------------------- Predictive Models for result displays:
    def predictive_individual_cov(self, precision_within):
        """Covariance of the Predictive distribution of IRT traits
        for a RANDOM INDIVIDUAL
        in the sub-population represented by this subject group.
        :param precision_within: TraitPrecisionWithinGroup object defining
            current distribution of inter-individual precision matrix within every group
        :return: 2D array with covariance matrix

        Method: The exact predictive distribution is a continuous mixture
        of Student-t distributions.
        Samples from the exact distribution can be generated,
        but it may be sufficient to know the exact covariance
        derived from self.precision_within and self.precision_among
        The predictive location is simply = self.mu.loc
        """
        return precision_within.mean_inv + self.mu.cov

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.theta and self.mu pruned
        """
        th = self.theta[..., keep]
        self.n_traits = np.sum(keep)
        self.theta = th
        self.mu.prune(keep)

    def standardize(self, s):
        """Standardize for unity trait variance, as estimated externally
        :param s: 1D array with scale factors, one for each trait
            s.shape[-1] == self.theta.shape[-1]
        :return: None
        Method: All trait samples down-scaled by s
        """
        self.theta = self.theta / s
        self.mu.standardize(s)

    def mean_response(self):
        """Mean rating across all items, disregarding any missing data
        used mainly to illustrate deviation re. corresponding mean trait
        :return: mean_r = 1D array
        """
        r = self.subject_responses
        # r[i, s] = integer ordinal response to i-th item by s-th subject
        # with origin 1, missing response = 0
        mean_r = np.sum(r, axis=0) / np.sum(r > 0, axis=0)
        return mean_r

    def scaled_subject_responses(self, item_scale_values):
        """Recoded subject responses on common interval scale
        :param item_scale_values: nested list with point-estimated scale values
            item_scale_values[i][r] = value for r-th response to i-th item
            assuming scale mean = 0, to be used for missing responses
        :return: sc_r = 2D array
            sc_r[i, s] = recoded response value for i-th item by s-th subject
            with missing responses recoded as zero
        """
        r = self._response
        # r[i, s] integer-encoded with origin=0, missing response == -1
        sc_r = np.array([v_i[np.maximum(0, r_i)]
                         for (v_i, r_i) in zip(item_scale_values, r)])
        sc_r[r < 0] = 0.
        return sc_r

    def predictive_mean(self):
        """Predictive distribution of MEAN IRT traits
        in the sub-population for which the subject data in self are representative.
        :return: GroupMeanTrait distribution object
        """
        return self.mu

    def predictive_individual(self, precision_within):
        """Predictive distribution of IRT traits for a RANDOM INDIVIDUAL
        in the sub-population for which the subject data in self are representative.
        :param precision_within: distribution instance for trait precision, given mean
        :return: PredictiveTrait object
        """
        return GaussianGivenParam(mu=self.mu, tau=precision_within)


# -------------------------------------------------------------------------
class GroupMeanTrait:
    """Gaussian distribution of group mean trait vector mu,
    defined by properties
    loc = location, learned from sampled trait vectors theta of ONE group of subjects,
    prec = precision matrix, learned externally by data from all groups
    """
    def __init__(self, loc, prec=None):
        self.loc = np.asarray(loc)
        if prec is None:
            self.prec = np.eye(len(loc))
        else:
            self.prec = np.asarray(prec)

    @property
    def n_traits(self):
        return len(self.loc)

    @property
    def mean2(self):
        """Mean square elements E{ mu_d^2 | self }
        used externally for learning precision_among_groups property
        for case with NON-tied precision_among_groups
        :return: mean2 = 1D array
        """
        return self.loc**2 + np.diag(self.cov)

    @property
    def cov(self):
        return np.linalg.inv(self.prec)

    def adapt(self, mean_theta, mean_prec_within, prec_among):
        """Adapt self to observed samples
        :param mean_theta: 2D array of mean observed samples of subject traits
            mean_theta[s, d] = current mean of d-th trait for s-th subject in this group
        :param mean_prec_within: mean precision matrix within this group
        :param prec_among: TraitPrecisionAmongGroups object
        :return: LL = log-likelihood after adaptation
        """
        (n_subjects, n_traits) = mean_theta.shape
        psi_among = np.eye(n_traits) * prec_among.mean  # ** works also for NON-tied prec_among
        self.prec = n_subjects * mean_prec_within + psi_among
        sum_theta = np.sum(mean_theta, axis=0)
        self.loc = np.linalg.solve(self.prec,
                                   np.dot(mean_prec_within, sum_theta))
        # = inv(self.prec) @ mean_prec_within @ sum_theta
        kl_div = self.relative_entropy(prec_among)
        # kl_div_tied = self.relative_entropy_tied(prec_among)  # ******** only for test
        # logger.info(f'group.mu.kl_div_tied= {kl_div_tied:.3f}; kl_div= {kl_div:.3f}')
        return - kl_div

    def relative_entropy(self, prior_prec):
        """Mean KLdiv( self || prior), where
        prior is zero-mean Gaussian vector with precision defined by given prior_prec,
        with either TIED or NON-TIED diagonal prec elements
        :param prior_prec: TraitPrecisionAmongGroups object defined externally
        :return: E{ ln pdf(mu | self) / pdf(mu | prior_prec) }_{self, prior_prec}

        Method:
        pdf(mu | prior) = prod_d sqrt(psi_d / 2 pi) exp(-mu_d^2 psi_d / 2), where
        psi is precision diagonal defined by prior_prec, either scalar or vector-valued
        pdf(mu | self) = ( |Psi|^{1/2} / (2 pi)^{D/2} ) exp(-(mu - m)^T Psi (mu - m) / 2
        where Psi = self.prec; m = self.loc

        KLdiv = ( ln(|Psi|) - sum_d(E{ln(psi_d)}) )/2 - D/2 + sum_d( E{ mu_d^2 } * E{psi_d}) /2
        """
        # d = self.n_traits
        (s, log_det_prec) = np.linalg.slogdet(self.prec)
        return 0.5 * (s * log_det_prec
                      + np.sum(self.mean2 * prior_prec.mean - prior_prec.mean_log - 1)
                      )

    def rvs(self, size):
        """Generate random samples
        :param size: scalar number of samples
        :return: 2D array with
            mu[n, t] = n-th sample of t-th trait
        """
        return multivariate_normal(mean=self.loc, cov=self.cov, size=size)

    def prune(self, keep):
        """Prune away un-necessary trait dimensions
        :param keep: 1D boolean array indicating traits to keep
        :return: None
        Result: self.prec and self.loc pruned
        """
        self.loc = self.loc[keep]
        self.prec = self.prec[keep, :][:, keep]

    def standardize(self, s):
        """Standardize by scaling for unity trait variance, as estimated externally
        :param s: 1D array with down-scale factors, one for each trait
        :return: None
        Method: All trait samples to be divided by s
            loc down-scaled, prec up-scaled, i.e., cov downscaled
        """
        self.loc = self.loc / s
        self.prec = self.prec * s * s[:, None]


# ---------------------------------------- module help functions

def item_logprob_by_theta(theta, tau_i, w_i, r_i):
    """sum log prob of observed data for ONE item,
    averaged across tau samples
    :param theta: theta[n, s, t] = n-th sample of t-th trait by s-th subject
    :param tau_i: tau_i[m, l] = m-th sample of UPPER interval limit for l-th response
        tau_i.shape[-1] == L_i - 1
    :param w_i: 1D array with w_i[t] = i-th item weight for t-th trait
    :param r_i: r_i[s] = response index for s-th subject to i-th item
    :return: lp = 1D array, with lp[n] = n-th sample
    """
    (tau_low, tau_high) = response_interval(tau_i, r_i)
    # tau_low[m, s] and tau_high[m, s] are (low, high) interval limits
    th = theta[:, None, :, :]  # to allow broadcast
    # th[n, 0, s, t] = n-th sample of t-th trait for s-th subject
    lp = log_prob_range(tau_low[..., None] - th, tau_high[..., None] - th)
    # lp[n, m, s, t] = n-th sample of t-th trait for s-th subject, with m-th tau sample
    lp = np.sum(lp, axis=-2)  # across subjects
    # lp[n, m, t]
    lp = np.mean(lp, axis=-2)
    # lp[n, t]
    lp = np.dot(lp, w_i)  # ***** let caller do this ???
    # lp[n]
    return lp


def d_item_logprob_by_theta(theta, tau_i, w_i, r_i):
    """d_logprob_by_theta of response data for each item
    :param theta: theta[n, s, t] = n-th sample of t-th trait by s-th subject
    :param tau_i: tau_i[m, l] = m-th sample of UPPER interval limit for l-th response
        tau_i.shape[-1] == L_i - 1
    :param w_i: 1D array with w_i[t] = i-th item weight for t-th trait
    :param r_i: list with r_i[s] = response index for s-th subject to i-th item
    :return: generator yielding dlp = 2D array, with
        dlp[n, s] = n-th sample of gradient for s-th subject theta value
    """
    (tau_low, tau_high) = response_interval(tau_i, r_i)
    # tau_low[m, s] and tau_high[m, s] are (low, high) interval limits
    th = theta[:, None, :, :]  # to allow broadcast
    # th[n, 0, s, t] = n-th sample of t-th trait for s-th subject
    dlp = d_log_prob_range_dm(tau_low[..., None] - th, tau_high[..., None] - th)
    # dlp[n, m, s, t] = n-th theta sample, m-th tau sample for s-th subject
    dlp = np.mean(dlp, axis=1)
    # dlp[n, s, t] = n-th sample of t-th trait for s-th subject
    return dlp * w_i    # ***** let caller do the weighting ???


def response_interval(tau_i, r_i):
    """Get response interval for given thresholds and responses for ONE item
    :param tau_i: 2D array with threshold samples, as
        tau_i[m, l] = m-th sample of UPPER limit of l-th response interval,
            EXCEPT extreme limits -infty, +infty
        tau_i.shape[-1] == L - 1, with L = highest response value
    :param r_i: 1D array of response indices, with missing response = -1
        r_i[s] = response index for s-th subject in range(L)
        except r_[s] = -1 in case of missing response 
    :return: tuple (tau_low, tau_high) with
        tau_low[m, s] = m-th sample of lower interval limit for r_i[s]
        tau_high[m, s] = m-th sample of upper interval limit for r_i[s]
        tau_low.shape == tau_high.shape
    """
    r_i = np.asarray(r_i)
    max_l = tau_i.shape[-1] - 1
    # r_i may be -1 for missing response, or = max_l +1 for highest interval
    tau_low = tau_i[:, np.maximum(0, r_i - 1)]
    tau_low[:, r_i <= 0] = -np.inf  # lowest interval OR missing response value
    tau_high = tau_i[:, np.minimum(max_l, r_i)]
    tau_high[:, r_i > max_l] = np.inf  # highest interval
    tau_high[:, r_i < 0] = np.inf  # missing response value
    return tau_low, tau_high


def log_prob_range(a, b):
    """log prob{ a < R <= b },
    for standard logistic random variable R, i.e.,
    = log(logistic.cdf(b) - logistic.cdf(a))
    :param a: array of lower interval limits
    :param b: array of upper interval limits, such that
        a.shape == b.shape
        -inf <= a < b <= +inf
    :return: lP = array of logprob values
        lP.shape == a.shape
    """
    # Reformulated, numerically safer than use of cdf functions:
    return (np.log1p(-np.exp(a - b))
            - np.log1p(np.exp(-b))
            - np.log1p(np.exp(a))
            )


def d_log_prob_range_dm(am, bm):
    """Gradient d log_prob_range(am, bm) / dm, where
    am = a - m, and bm = b - m
    :param am: array with lower limits
    :param bm: array with upper limits
        am.shape == bm.shape
        -inf <= am  < bm <= +inf
    :return: d_dm = array of gradient elements
        d_dm.shape == am.shape == bm.shape

    Method:
    log_prob_range = log-prob( a < Z <= b),
    for a logistic-distributed Z with location = m and scale=1,
    = logprob(a-m < Y <= b-m) for standardized Y with location = 0.
    = ln(1 - e^(am-bm)) - ln(1 + e^(-bm)) - ln(1 + e^(am))
    am - bm = a - b is independent of m.
    Thus,
    d_dm = - e^{-bm) / (1+ e^{-bm} ) + e^am / (1 + e^am)
    """
    return logistic.cdf(am) - logistic.cdf(-bm)


def d_log_prob_range_dab(am, bm):
    """Gradient d log_prob_range(am, bm) w.r.t. interval limits (a, b), where
    am = a - m, and bm = b - m
    :param am: array with lower limits
    :param bm: array with upper limits
        am.shape == bm.shape
        -infty <= am  < bm <= +infty
    :return: tuple (d_da, d_db), where
        d_da = d logprob(am, bm) w.r.t lower limit a
        d_db = d logprob(am, bm) w.r.t upper limet b
        d_da.shape == d_db.shape == am.shape == bm.shape

    Method:
    log_prob_range = log-prob( a < Z <= b),
    for a logistic-distributed Z with location = m and scale=1,
    = logprob(a-m < Y <= b-m) for standardized Y with location = 0.
    = ln(1 - e^(am-bm)) - ln(1 + e^(-bm)) - ln(1 + e^(am))

    d_da = -1 / (e^{bm - am} - 1 ) - e^am / (1 + e^am)
    d_db = +1 / (e^{bm - am} - 1 ) + e^bm / (1 + e^bm)
    """

    c = 1 / np.expm1(bm - am)
    if np.any(np.logical_not(np.isfinite(c))):
        logger.warning('Invalid tau, zero interval width')
    return - c - logistic.cdf(am), c + logistic.cdf(-bm)


# -------------------------------------------------------- TEST:
if __name__ == '__main__':
    # Testing some module functions

    from scipy.optimize import check_grad, approx_fprime

    print('\n*** Check logprob_range gradients: ')
    a = np.array([5.])
    b = a + 2.

    def test_logprob_range_m(y):
        return log_prob_range(a- y, b- y)

    def test_d_logprob_range_dm(y):
        return d_log_prob_range_dm(a- y, b- y)

    # ------------------------------------
    def test_logprob_range_a(y):
        return log_prob_range(a + y, b)

    def test_d_logprob_range_da(y):
        return d_log_prob_range_dab(a + y, b)[0]

    # ----------------------------------------
    def test_logprob_range_b(y):
        return log_prob_range(a, b + y)

    def test_d_logprob_range_db(y):
        return d_log_prob_range_dab(a, b + y)[1]

    test_y = np.array([0.])
    print(f'test_logprob_range_m(test_y) = {test_logprob_range_m(test_y)}')

    err = check_grad(test_logprob_range_m, test_d_logprob_range_dm, test_y)
    print('d_logprob_range_dm =', test_d_logprob_range_dm(test_y))
    print('approx_grad = ', approx_fprime(test_y,
                                          test_logprob_range_m,
                                          epsilon=1e-6))
    print('check_grad err = ', err)

    print(f'test_logprob_range_a(test_y) = {test_logprob_range_a(test_y)}')
    err = check_grad(test_logprob_range_a, test_d_logprob_range_da, test_y)
    print('d_logprob_range_da =', test_d_logprob_range_da(test_y))
    print('approx_grad = ', approx_fprime(test_y,
                                          test_logprob_range_a,
                                          epsilon=1e-6))
    print('check_grad err = ', err)

    print(f'test_logprob_range_b(test_y) = {test_logprob_range_b(test_y)}')
    err = check_grad(test_logprob_range_b, test_d_logprob_range_db, test_y)
    print('d_logprob_range_db =', test_d_logprob_range_db(test_y))
    print('approx_grad = ', approx_fprime(test_y,
                                          test_logprob_range_b,
                                          epsilon=1e-6))
    print('check_grad err = ', err)

    # ---------------------------------------------------------------
    print('\n*** Testing ResponseGroup ***\n')
    n_samples = 1
    n_subjects = 2
    n_traits = 3
    n_items = 3

    theta = np.array([[1., 2., 3.], [0., -1., -2.]])  # 2 subjects, 3 traits
    theta = np.tile(theta, (n_samples, 1, 1))

    s_response = np.array([[1, 2], [3, 4],[0, 4]]) + 1  # 3 items, 5 response levels
    # encoded as 1,...,5
    trait_item_map = np.eye(n_traits, dtype=bool)
    mu = GroupMeanTrait(loc=np.zeros(n_traits))

    g_ind = ResponseGroup(s_response,
                          theta=theta,
                          mu=mu,
                          name='TEST')
    # NOTE: response encoded with origin == 1 here!  **********************
    print('g_ind=', g_ind)
    print('g_ind.theta=\n', g_ind.theta)
    print('g_ind._response=\n', g_ind._response)
    print('g_ind.subject_responses=\n', g_ind.subject_responses)

    n_tau_samples = 3
    tau = [np.array([[-2., -1., +2., +4]
                     for _ in range(n_tau_samples)])
           for _ in range(3)]  # 3 items, 5 response levels
    print('tau= ', tau)
    r_i = np.arange(-1, 5, dtype=int)
    (a, b) = response_interval(tau[0], r_i)
    for (l, a_l, b_l) in zip(r_i, a[0], b[0]):
        print(f'\tresponse_interval a,b({l})= {a_l, b_l}')

    #  *** th = theta.reshape((n_samples, -1))
    print('sampler theta= ', theta)
    weight = np.eye(n_traits)  # ********** item_i <=> trait_i
    # weight = np.ones((n_items,n_traits))  # ********** item_i <=> trait_i

    lp = g_ind.logprob_by_theta(theta, tau, weight)
    print('logprob_by_theta(theta, tau, weight)= ', lp)

    mean_prec_within = np.eye(n_traits)
    mean_log_det_prec = 1.
    lp = g_ind.prior_logprob(theta, mean_prec_within, mean_log_det_prec)
    print('prior_lp(theta, tau)= ', lp)

    th = theta.reshape((n_samples, -1))
    lp = g_ind.neg_logprob(th, tau, weight, mean_prec_within, mean_log_det_prec)
    print('neg_logprob(th, tau)= ', lp)

    # -----------------------------------------------
    def test_neg_logprob(y):
        return g_ind.neg_logprob(y.reshape(1,-1), tau, weight, mean_prec_within, mean_log_det_prec)[0]

    def test_grad_neg_logprob(y):
         return g_ind.d_neg_logprob(y.reshape(1,-1), tau, weight, mean_prec_within, mean_log_det_prec)[0]
    # -------------------------------------------------
    print('test_neg_logprob(th)= ', test_neg_logprob(th))

    y = th[0]  # must be 1D vector for test
    err = check_grad(test_neg_logprob, test_grad_neg_logprob, y)
    print('grad_test =', test_grad_neg_logprob(y))
    print('approx_grad = ', approx_fprime(y,
                                          test_neg_logprob,
                                          epsilon=1e-6))
    print('check_grad err = ', err)

    print('\nTesting item_logprob_by_tau\n')

    item_ind = 0
    w_i = weight[item_ind]

    # ----------------------------------------
    def test_logprob_mean_theta(y):
        tau = y[None, :]
        return g_ind.item_logprob_by_tau(tau, item_ind, w_i)[0]

    def test_d_logprob_mean_theta(y):
        tau = y[None, :]
        return g_ind.d_item_logprob_by_tau(tau, item_ind, w_i)[0]

    test_tau = np.array(tau)[0,0,:]  # must be vector for test
    print('test_tau= ', test_tau)
    print('test_logprob_mean_theta(test_tau, item_ind)= ', test_logprob_mean_theta(test_tau))

    err = check_grad(test_logprob_mean_theta, test_d_logprob_mean_theta, test_tau)
    print('grad_test =', test_d_logprob_mean_theta(test_tau))
    print('approx_grad = ', approx_fprime(test_tau,
                                          test_logprob_mean_theta,
                                          epsilon=1e-6))
    print('check_grad err = ', err)

    test_tau = np.array(tau)[0, 0, :]  # must be vector for test
    print('test_tau= ', test_tau)
    print('test_logprob_mean_theta(test_tau, item_ind)= ', test_logprob_mean_theta(test_tau))

    err = check_grad(test_logprob_mean_theta, test_d_logprob_mean_theta, test_tau)
    print('grad_test =', test_d_logprob_mean_theta(test_tau))
    print('approx_grad = ', approx_fprime(test_tau,
                                          test_logprob_mean_theta,
                                          epsilon=1e-6))
    print('check_grad err = ', err)



"""This module defines data structures and functions to display analysis results
given an OrdinalItemResponseModel instance,
learned from one or several data sets with ordinal responses to a questionnaire.

Results are shown as figures and tables.
Figures can be saved in pdf, eps, or other formats, as specified in ir_display_format.
Tables are saved in LaTeX tabular format OR in tab-delimited text files.
Thus, both figures and tables can be easily imported into a LaTeX document
or other word-processing document.

*** Main Class:

ItemResponseDisplaySet = a structured container for all display results
Each display element can be accessed and modified by the user, before saving.

*** Usage Example: see template script run_irt

*** Version History:
2020-06-04, first version for ItemResponseCalc standalone package
"""
# ********** figure comments table comments

import numpy as np
from collections import Counter
from pathlib import Path

import logging

from samppy.credibility import cred_diff

from . import ir_display_format as fmt

# -------- module-global variables, that may be changed by user:
FMT = {'percentiles': [5., 50., 95.],
       'credibility_levels': [.99, .95, .9, .8, .7],  # MAX 5 values
       'population': 'Population',
       'trait': 'Trait',
       'mean': 'Mean',
       'individual': 'Indiv.',
       'equivalent': 'Equivalent',
       'response': 'Response',
       'probability': 'Prob.',
       'rating': 'Rating',
       'global': 'Global',
       'correlation': 'Correlation',
       }
# = dict with format parameters that may be changed by user
# Any other parameters are handled by module ir_display_format

logger = logging.getLogger(__name__)
# logger.setLevel(logging.INFO)
# logger.setLevel(logging.DEBUG)  # test


# ------------------------------------------------------------
def set_format_param(**kwargs):
    """Set / modify format parameters for this module
    :param kwargs: dict with any formatting variables
    :return: None
    """
    other_fmt = dict()
    for (k, v) in kwargs.items():
        k = k.lower()
        if k in FMT:
            FMT[k] = v
        else:
            other_fmt[k] = v
    if len(FMT['credibility_levels']) > 5:
        FMT['credibility_levels'] = FMT['credibility_levels'][:5]
    fmt.set_format_param(**other_fmt)  # all remaining user-defined parameters


# ---------------------------------------------------------- Main Display Classes:
class ItemResponseDisplaySet:
    """Root container for all displays of predictive trait results
    and item properties,
    from one OrdinalItemResponseModel object.
    All display elements can be saved as files within a selected directory three.
    The complete instance can also be serialized and dumped to a single pickle file,
    then re-loaded and re-saved, if any display object needs to be edited.
    """
    def __init__(self,
                 log_likelihood,
                 instrument,
                 traits):
        """
        :param log_likelihood: fmt.FigureRef with log-likelihood plot
        :param instrument: ItemSetDisplay object, showing
            response thresholds, and response_prob plot for each item
        :param traits: TraitDisplay object with predicted trait distribution
            for population mean, and for random individual in population.
            Traits for individual respondents are NOT (yet) displayed.
        """
        self.log_likelihood = log_likelihood
        self.instrument = instrument
        self.traits = traits

    def __repr__(self):
        skip = '\n\t'
        return (self.__class__.__name__ + '(' + skip +
                (',' + skip).join(f'{key}={repr(v)}'
                                  for (key, v) in vars(self).items()) +
                skip + ')')

    def save(self, dir_top):
        """Save all displays in a directory tree
        :param dir_top: Path or string with top directory for all displays
        :return: None
        """
        dir_top = Path(dir_top)  # just in case
        self.log_likelihood.save(dir_top)
        if self.instrument is not None:
            self.instrument.save(dir_top)
        if self.traits is not None:
            self.traits.save(dir_top)

    @classmethod
    def show(cls, irm, mapping_item=-1, **kwargs):
        """Create displays for all results from an Item Response analysis model,
        and store all display elements in a single structured object.
        :param irm: a single learned OrdinalItemResponseModel instance
        :param mapping_item: index of irm.scales for item to be used for
            mapping non-linear mean rating, given scale trait
        :param: kwargs: (optional) any display formatting parameters
        :return: a single cls instance
        """
        set_format_param(**kwargs)
        logger.info(f'Creating displays for {len(irm.groups)} group(s): ' +
                    ', '.join(g.name for g in irm.groups))

        return cls(log_likelihood=fmt.fig_log_likelihood(irm.log_prob,
                                                         irm.n_subjects,
                                                         color='r',
                                                         name='LearnedLL'),
                   instrument=ItemSetDisplay.show(irm),
                   traits=TraitDisplay.show(irm, mapping_item))


class ItemSetDisplay:
    """Container for all displays of item-related properties
    """
    def __init__(self,
                 response_freq,
                 items):
        """
        :param response_freq: FigureRef with relative response frequencies, all items
        :param items: dict with elements (item_id, item_display),
            item_display = ItemDisplay object
        """
        self.response_freq = response_freq
        self.items = items

    def save(self, top_path):
        """
        :param top_path:
        :return: None
        """
        items_path = top_path / 'items'
        if self.response_freq is not None:
            self.response_freq.save(items_path)
        for item in self.items.values():
            item.save(items_path)

    @classmethod
    def show(cls, irm):
        """
        :param irm: OrdinalItemResponseModel instance
        :return: a cls instance
        """
        t_min = min(np.min(s.tau) - 2. * s.latent_scale
                    for s in irm.scales)
        t_max = max(np.max(s.tau) + 2. * s.latent_scale
                    for s in irm.scales)
        t = np.linspace(t_min, t_max, 100)
        # same trait range across all items
        g_count = irm.item_response_count()
        # = dict with list of Counter objects, one for each item
        total_count = [sum((g_c[i] for g_c in g_count.values()), Counter())
                       for i in range(irm.n_items)]
        # *** use item_id instead of item_index ********************
        return cls(response_freq=fmt.fig_response_freq(total_count,
                                                       name='Response_freq'),
                   items={f'Q{scale.item_index + 1}': ItemDisplay.show(scale,
                                                                       count,
                                                                       trait=t)
                          for (scale, count) in zip(irm.scales,
                                                    total_count)})


class ItemDisplay:
    """Container for all displays related to ONE item
    """
    def __init__(self,
                 response_freq=None,
                 response_prob=None,
                 mean_response=None):
        """
        :param response_freq: FigureRef with empirical response freq
        :param response_prob: FigureRef with predicted response prob vs. trait
        :param mean_response: FigureRef with expected response vs. trait
            like "Item Characteristic Curves" in the Rasch tradition.
            (*** and ? empirical response range vs trait ? *****)
            NOTE: NOT valid, as ordinal responses are NOT on interval scale
            SKIPPING this display variant for now...
        """
        self.response_freq = response_freq
        self.response_prob = response_prob
        self.mean_response = mean_response

    def save(self, save_path):
        """
        :param save_path: path where sub-objects are saved
        :return:
        """
        if self.response_freq is not None:
            self.response_freq.save(save_path)
        if self.response_prob is not None:
            self.response_prob.save(save_path)
        if self.mean_response is not None:
            self.mean_response.save(save_path)

    @classmethod
    def show(cls, item_scale, response_count, trait):
        """Generate displays for given item
        :param item_scale: OrdinalItemScale object
        :param response_count: Counter object with response counts
        :param trait: 1D array with dense trait values for plot
        :return: a cls instance
        """
        item_id = f'Q{1+item_scale.item_index}'
        # ***** use item_scale.item_id instead *********************
        tau = np.mean(item_scale.tau, axis=0)
        ord_prob = item_scale.ordinal_prob(trait)
        r_mean = 1. + np.dot(ord_prob, np.arange(item_scale.n_response_levels))
        # response_freq=fmt.fig_response_freq(response_count),
        return cls(response_prob=fmt.fig_response_prob(trait, ord_prob,
                                                       name=item_id + '_prob',
                                                       tau=tau,
                                                       x_label=FMT['trait'],
                                                       y_label=' '.join([FMT['response'],
                                                                         FMT['probability']])
                                                       ))


class TraitDisplay:
    """Container for result displays of predicted trait values
    for all traits and groups,
    with layout depending on number of groups and traits.
    """
    def __init__(self, pop_mean, pop_ind,
                 corr_global=None, corr_within=None,
                 group_diff=None):
        """
        :param pop_mean: FigureRef of population predictive mean traits
        :param pop_ind: FigureRef of population predictive traits for random individual
        :param corr_global: (optional) TableRef with global predictive trait correlations
        :param corr_within: (optional) TableRef with predictive trait correlations
            within populations, i.e., excl. variance between populations.
        :param group_diff: (optional) GroupDiffDisplay object,
            iff we have more than one group (population)
       """
        self.pop_mean = pop_mean
        self.pop_ind = pop_ind
        self.corr_global = corr_global
        self.corr_within = corr_within
        self.group_diff = group_diff

    def save(self, path):
        if self.pop_mean is not None:
            self.pop_mean.save(path)
        if self.pop_ind is not None:
            self.pop_ind.save(path)
        if self.corr_global is not None:
            self.corr_global.save(path)
        if self.corr_within is not None:
            self.corr_within.save(path)
        if self.group_diff is not None:
            for (trait_id, trait_diff) in self.group_diff.items():
                trait_diff.save(path / 'group_diff', name=trait_id)

    @classmethod
    def show(cls, irm, mapping_item=-1):
        """
        :param irm: OrdinalItemResponseModel instance
        :param mapping_item: index of item to be used for non-linear scale mapping
        :return: a cls instance
        """
        def _fig_percentiles(perc, g_labels, t_labels, y_label):
            """Plot trait percentiles vs groups or traits,
            in style depending on n_groups and n_traits
            :param perc: 3D array with percentile trait values
                perc[g, p, t] = p-th percentile of t-th trait in g-th population
            :param g_labels: list of group label strings
                len(g_labels) == perc.shape[0]
            :param t_labels: list of trait label strings
                len[t_labels] == perc.shape[2]
            :param y_label: string with y-axis label
            :param mapping_item: index of irm.scales to be used for mapping
            :return: FigureRef instance
            """
            (n_groups, n_perc, n_traits) = perc.shape
            if n_traits > 1:  # plot perc vs traits, with populations as sub-cases
                f = fmt.fig_percentiles(perc,
                                        x_label=FMT['trait'],
                                        x_ticklabels=t_labels,
                                        case_labels=g_labels,
                                        y_label=y_label,
                                        x_offset=0.07)
            else:  # plot perc vs groups, with traits as sub-cases
                f = fmt.fig_percentiles(perc.transpose(2, 1, 0),
                                        x_label=FMT['population'],
                                        x_ticklabels=g_labels,
                                        case_labels=t_labels,
                                        y_label=y_label,
                                        x_offset=0.07)
            # ax = f.ax
            # make second corresponding scale with expected response ratings
            (y_min, y_max) = f.ax.get_ylim()
            y_min -= 0.1  # **** fix to make more space for legend
            f.ax.set_ylim(y_min, y_max)
            y = np.linspace(y_min, y_max, 100)
            # use mapping_item and its item_id
            y_mapped = 1. + irm.scales[mapping_item].mean_ordinal(y)
            item_id = f'Q{1+irm.scales[mapping_item].item_index}'
            # item_id = irm.scales[mapping_item].item_id
            f.mapped_y_axis(y, y_mapped,
                            y_label=(FMT['equivalent'] + ' ' +
                                     item_id + ' ' +
                                     FMT['rating'])
                            )
            return f
        # -----------------------------------------------------------
        item_trait_map = irm.item_trait_map()
        trait_sel = np.any(item_trait_map, axis=0)
        # = boolean trait index for display, just in case irm has NOT been pruned
        item_trait_map = item_trait_map[:, trait_sel]
        n_traits = np.sum(trait_sel)
        group_labels = [g.name for g in irm.groups]
        group_mean_samples = np.array([g.predictive_mean().rvs(size=10000)[:, trait_sel]
                                       for g in irm.groups])
        group_mean_perc = np.percentile(group_mean_samples,
                                        FMT['percentiles'],
                                        axis=1)
        # group_mean_perc[p, g, t] = p-th percentile for t-th trait in g-th population
        trait_labels = fmt._make_trait_labels(item_trait_map)  # **** done by model?
        pop_mean = _fig_percentiles(group_mean_perc.transpose(1, 0, 2),
                                    group_labels,
                                    trait_labels,
                                    y_label=' '.join([FMT['population'],
                                                      FMT['mean'],
                                                      FMT['trait']])
                                    )
        group_ind = [np.percentile(g.predictive_individual(irm.precision_within).rvs(size=10000)[:, trait_sel],
                                   FMT['percentiles'],
                                   axis=0)
                     for g in irm.groups]
        pop_ind = _fig_percentiles(np.array(group_ind),
                                   group_labels,
                                   trait_labels,
                                   y_label=' '.join([FMT['population'],
                                                     FMT['individual'],
                                                     FMT['trait']])
                                   )

        c_global = None
        c_within = None
        if n_traits > 1:
            if irm.n_groups > 1:
                c_global = fmt.tab_correlation_matrix(irm.predictive_individual_cov(),
                                                      item_trait_map,
                                                      name='_'.join((FMT['trait'],
                                                                     FMT['correlation'],
                                                                     FMT['global'])
                                                                    ))
            c_within = fmt.tab_correlation_matrix(irm.precision_within.mean_inv,
                                                  item_trait_map,
                                                  name='_'.join((FMT['trait'],
                                                                 FMT['correlation'],
                                                                 FMT['population'])
                                                                ))

        g_diff = None
        if irm.n_groups > 1:
            g_diff = {trait_id: GroupDiffDisplay.show(t_samples,
                                                      group_labels)
                      for (trait_id, t_samples) in zip(trait_labels,
                                                       group_mean_samples.transpose(2, 0, 1))
                      }
            if n_traits > 1:
                g_diff['Mean_Trait'] = GroupDiffDisplay.show(np.mean(group_mean_samples,
                                                                     axis=-1),
                                                             group_labels)

        return cls(pop_mean=pop_mean,
                   pop_ind=pop_ind,
                   corr_global=c_global,
                   corr_within=c_within,
                   group_diff=g_diff)


class GroupDiffDisplay:
    """Container for displays showing predictive credible differences
    between populations, for ONE selected trait.
    """
    def __init__(self, fig=None, tab=None):
        """
        :param fig: FigureRef object
        :param tab: TableRef object
        """
        self.fig = fig
        self.tab = tab

    def save(self, path, name):
        """
        :param path: path to directory for saved plots and tables
        :param name: string with file name
        :return: None
        """
        if self.fig is not None:
            self.fig.save(path, name=name)
        if self.tab is not None:
            self.tab.save(path, name=name)

    @classmethod
    def show(cls, group_samples, group_labels):
        """Display credible differences between groups,
        given samples of predictive distribution
        :param group_samples: 2D array with
            group_samples[g, n] = n-th sample of selected trait for g-th group
        :param group_labels: list of string-valued group labels
            len(group_labels) == group_samples.shape[0]
        :return: a cls instance
        """
        # **** Sort groups for decreasing median values *************
        s_ind = np.argsort(np.median(group_samples, axis=1))[::-1]
        g_samples = group_samples[s_ind, :]
        g_labels = [group_labels[i] for i in s_ind]
        diff = cred_diff(g_samples,
                         diff_axis=0, sample_axis=1,
                         p_lim=np.amin(FMT['credibility_levels']))

        fig = fmt.fig_credible_diff(diff,
                                    x_labels=g_labels,
                                    x_label=FMT['population'],
                                    cred_levels=FMT['credibility_levels'])
        tab = fmt.tab_credible_diff(diff,
                                    x_labels=g_labels,
                                    x_label=FMT['population'])
        return cls(fig, tab)



"""Main template script to analyze ordinal response data using Item Response Theory.
This template should be copied and edited by the user for the desired application.

Arne Leijon, 2020-05-30, template example based on IOI-HA data analysis
2020-06-06, standalone version for ItemResponseCalc package
"""

# -------- __main__ check to prevent multiprocessor sub-tasks to re-run this script
if __name__ == '__main__':

    import numpy as np
    import pickle
    from pathlib import Path
    import datetime as dt
    import logging

    from ItemResponseCalc import ir_logging
    from ItemResponseCalc.item_response_data import Questionnaire
    from ItemResponseCalc.item_response_data import ItemResponseDataSet
    from ItemResponseCalc.item_response_data import item_response_table
    from ItemResponseCalc.table_reader import Table, Tables
    # Tables used to join data in teo or more files into one group
    from ItemResponseCalc.item_response_model import OrdinalItemResponseModel
    from ItemResponseCalc.ir_display import ItemResponseDisplaySet

    # ----------------------------------------------- ItemResponseModel parameters:
    max_n_traits = 4
    # = max number of traits to explain total set of item responses
    # = 1, if the test instrument is known to measure a uni-dimensional trait

    # n_scale_samples = 50
    # -> sampled set of scale threshold vectors
    n_scale_samples = 1
    # -> MAP-estimated single scale threshold vector
    n_subject_samples = 20
    # = number of sampled trait(s) for each respondent

    # NOTE: computation time is proportional to
    # n_subjects * n_scale_samples * n_subject_samples
    # n_scale_samples = 1 may give similar result as sampled scale model

    TEST_DOWNSAMPLE = 1
    # -> no down-sampling -> all records included
    # TEST_DOWNSAMPLE = 10
    # -> use smaller data subset only for faster initial test run

    # ------------------------------------------------------------------------
    TOP_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry'
    TOP_PATH = TOP_PATH / 'IRT-IOI-HA'
    # Change this example to top-dir for everything to be read or written here

    ioiha_data_path = TOP_PATH / 'IOI-HA-data'
    # path to actual item data files OTHER THAN data in sql server

    result_dir = TOP_PATH / 'IOI-HA-results'
    # path where results will be saved, for several runs if needed

    ioiha_file = ioiha_data_path / 'IOI-HA-English.txt'
    # text file containing the questions and response alternatives
    # If not available, a Questionnaire object must be created manually
    # as suggested below.

    # ------------------------------------------------------------------------
    timefmt = '{0.year}-{0.month:02}-{0.day:02}-{0.hour:02}{0.minute:02}'
    result_run = timefmt.format(dt.datetime.now())
    # -> name of sub-folder in result_dir

    save_dir = result_dir / result_run
    save_dir.mkdir(parents=True, exist_ok=True)
    # sub-path for results of this particular run

    model_file = 'irt-model.pkl'
    # = name of saved model in save_dir

    log_file = 'run_irt_log.txt'
    # = name of log file saved in same sub-folder
    # -------------------------------------------------------------------------

    # ------------------------------------- setup logging:
    ir_logging.setup(save_dir / log_file)
    logger = logging.getLogger(__name__)

    # -------------------------- common help functions for data input:
    def accept_record(r):
        """Inclusion criterion for accepting a record,
        to be called after any other data recoding in each input Table object.
        :param r: list with responses in ONE record read from file,
            encoded as integer values
        :return: boolean == True if record is acceptable
        """
        n_missing = sum(r_i is None or r_i < 1 or r_i > 5
                        for r_i in r)
        return n_missing <= 3

    def recode_missing_as_0(row):
        """Replace missing data by zero, as required by analysis.
        Table readers encode missing data as value = None
        :param row: list with responses in ONE record read from file
        :return: None, row elements recoded in place
        """
        for i in range(len(row)):
            if row[i] is None or row[i] < 1:  # or row[i] > 5: only for IOI-HA
                row[i] = 0
    # --------------------------------------------

    if TEST_DOWNSAMPLE > 1:
        logging.info(f'*** Test with down-sampling {TEST_DOWNSAMPLE}')

    # ------------------------------- Define questionnaire instrument
    # create a Questionnaire object to specify
    # number of items and response alternatives
    q_items = Questionnaire.load(ioiha_file)  # load from text file

    # if the text file is not available, create q_items manually, e.g.:
    # r_levels = [5, 5, 5, 5, 5, 5, 5]
    # # = list with number of response levels, one element for each item
    # # Example: all seven IOI-HA items have 5 response levels
    # q_items = Questionnaire(item_response_levels=r_levels)

    # ------------------------------- Define included data sets:
    data_groups = dict()  # space for all included groups

    # Example: Include two xlsx files from Hickson, Australia:
    # Responses have already been coded as integers 1,..., 5 in the file
    au1_file = ioiha_data_path / 'Hickson' / 'Short_Eartrak AUS.xlsx'
    logger.info(f'AU-H-10: Using {au1_file}')

    au1 = item_response_table(au1_file,
                              fields=[f'Q0{i}' for i in range(1, 8)],  # labels in header row
                              field_types=int,
                              missing_values=['.', ' '],
                              header_row=1,
                              sample_factor=TEST_DOWNSAMPLE,
                              recode_fcns=recode_missing_as_0,
                              accept_row=accept_record)
    data_groups['AU-H-10'] = au1

    au2_file = ioiha_data_path / 'Hickson' / 'Eartrakdata_10_May_08.xlsx'
    logger.info(f'AU-H-10e: Using {au2_file}')

    au2 = item_response_table(au2_file,
                              fields=[f'Q0{i}' for i in range(1, 8)],
                              field_types=int,
                              missing_values=['.', ' '],
                              header_row=1,
                              sample_factor=TEST_DOWNSAMPLE,
                              recode_fcns=recode_missing_as_0,
                              accept_row=accept_record)
    data_groups['AU-H-10e'] = au2

    # Include two csv files from S Kramer:

    nle_file = ioiha_data_path / 'Kramer' / 'moeder_final - kopieforArne.csv'
    logger.info(f'NL-02: Using {nle_file}')

    nle1 = item_response_table(file_path=nle_file,
                               fmt='csv',
                               fields=[f'ioi_v{i}' for i in range(1, 8)],
                               field_types=int,
                               missing_values=[9.0, ''],
                               sample_factor=TEST_DOWNSAMPLE,
                               recode_fcns=recode_missing_as_0,
                               accept_row=accept_record)

    # -------------------------------------------
    def recode_fields14(row):
        """IOI-HA items 1 and 4 coded as 0-4 in NLSH-Kramer-T2IOI-HA.csv
        must be recoded as values 1-5,
        BEFORE recoding missing-data as zero.
        """
        for i in [1, 4]:
            if row[i - 1] is not None:
                row[i - 1] += 1
    # --------------------------------------------

    nle2_file = ioiha_data_path / 'Kramer' / 'NLSH-Kramer-T2IOI-HA.csv'
    logger.info(f'NL-16: Using {nle2_file}')

    nle2 = item_response_table(file_path=nle2_file,
                               fmt='csv',
                               sample_factor=TEST_DOWNSAMPLE,
                               fields=[f'T2IOIHA{i + 1}' for i in range(7)],
                               missing_values=[9999999.0],
                               field_types=int,
                               recode_fcns=[recode_fields14,
                                            recode_missing_as_0],
                               # recode_fcns are called in given order
                               accept_row=accept_record)
    # Example: join both data files into ONE group:
    data_groups['NL-02-16'] = Tables(nle1, nle2)

    # *** Example: include data for one year from an SQL database:

    # from ItemResponseCalc.table_reader_sql import Table
    # = special Table reader for SQL access

    # haq_config = dict(user='user_id',
    #                   password='********',
    #                   host='127.0.0.1',  # IP address of SQL server
    #                   port='1111',  # access port to SQL server
    #                   database='hbdb',  # name of database on SQL server
    #                   connect_timeout=10  # seconds wait in case no access
    #                   )
    # = access arguments for SQL database

    # table = 'prescriptions'  # name of desired table in the database
    #
    # where = ("respondent = 1 " +
    #          "AND (Q2 = 1 OR Q2 = 2) " +
    #          "AND date BETWEEN %s AND %s"
    #          )
    # # = WHERE section of SELECT statement for SQL database
    # # with following two dates as arguments:
    # year_2019 = (dt.date(2019, 1, 1), dt.date(2019, 6, 30))

    # data_groups['SE-19'] = Table(table_id=table,
    #                              connect_kwargs=haq_config,  # SQL access
    #                              select_where=where,
    #                              select_args=year_2019,
    #                              fields=[f'Q{i}' for i in range(3, 10)],
    #                              # = IOI-HA field labels in database
    #                              recode_fcns=recode_missing_as_0,
    #                              accept_row=accept_record,
    #                              sample_factor=TEST_DOWNSAMPLE * 10)

    # --------------------------- Collect all groups into ONE DataSet object:

    ids = ItemResponseDataSet(questionnaire=Questionnaire.load(ioiha_file),
                              groups=data_groups)
    # NOT YET actually reading any data
    # In case of read error, it will show here:
    logging.info('Item Response Data collected = \n' + repr(ids))

    total_count = np.array([np.sum(c_i) for c_i in ids.item_response_count])
    logging.info('Total Response Counts: ' + np.array2string(total_count))

    # --------------------------- Create ItemResponseModel from data:

    irm = OrdinalItemResponseModel.initialize(ids,
                                              n_traits=max_n_traits,
                                              n_scale_samples=n_scale_samples,
                                              n_subject_samples=n_subject_samples,
                                              trait_scale=3.)  # *** = default

    logging.info(f'Learning model with max {irm.n_traits} traits'
                 + f' for {irm.n_groups} groups'
                 + f' with {irm.n_subjects} subject records in total')
    logging.info(f'The model uses {n_scale_samples} scale samples'
                 + f' and {n_subject_samples} samples for each subject')

    LL = irm.learn(max_hours=30)
    # *** It may take MANY hours, try setting shorter time limit first
    # When learning is complete, it will finish before the time limit
    # If learning was complete, the LL values should reach a plateau level

    logging.info(f'Learned model with {irm.n_groups} groups'
                 + f' with {irm.n_subjects} subjects in total')
    LL = np.array(LL)
    logging.info('Log Likelihood: ' + np.array2string(LL, precision=1))
    # if learning was complete, it should finish before the time limit,
    # and the LL values should should reach a nearly constant level

    # ------------------ save model for later input to show_irt script
    with (save_dir / model_file).open('wb') as f:
        pickle.dump(irm, f)

    logging.info(f'Result saved in {save_dir}')

    logging.info('Learned model = \n' + repr(irm))

    # ------------------------------------- Display Model Results:
    # The display can also be done by separate script show_irt.py

    # ------------------------------------- Trait Covariance before standardization
    # only informative, may be commented out
    c = irm.precision_within.mean_inv
    logging.info('Predictive Trait Covariance Within Populations =\n'
                 + np.array2string(c, precision=3, suppress_small=True))

    c = irm.precision_among.mean_inv
    logging.info('Predictive Trait Variance Among Populations= '
                 + np.array2string(c, precision=3, suppress_small=True))

    # ------------------------------------- Global Predictive Covariance
    c = irm.predictive_individual_cov()
    logging.info('Predictive Global Trait Cov (before standardization) =\n'
                 + np.array2string(c, precision=3, suppress_small=True))

    # -------------------------------------  prune and standardize:
    irm.prune()  # keep only traits that were really needed to model the data

    irm.standardize()  # re-scale all parameters for unity global trait variance

    c = irm.predictive_individual_cov()
    logging.info('Pruned and Standardized Predictive Global Trait Cov =\n'
                 + np.array2string(c, precision=3, suppress_small=True))
    s = np.sqrt(np.diag(c))
    logging.info('Predictive St.Dev. after Standardization = '
                 + np.array2string(s, precision=3))

    # ------------------------------------- display with standardized scale
    # mapping_item used to transform trait scale back to ONE item rating scale
    ir_display = ItemResponseDisplaySet.show(irm,
                                             mapping_item=-1,  # last questionnaire item
                                             figure_format='pdf',  # OR eps, png,...
                                             table_format='latex',  # OR tab
                                             )
    ir_display.save(save_dir)

    # ------------------------------------- Within-group trait correlation
    if irm.n_traits > 1:
        cov_within = irm.precision_within.mean_inv
        logging.info('Predictive Trait Covariance Within Populations:\n' +
                     np.array2string(cov_within, precision=3))
        s = np.sqrt(np.diag(cov_within))
        c = cov_within / s
        c /= s[:, np.newaxis]
        # = normalized correlation matrix, also tabulated in ir_display:
        logging.info('Predictive Trait Correlation Within Populations:\n' +
                     np.array2string(c, precision=3))

        # ------------------------------------- PCA of within-group covariance

        (eig_val, eig_vec) = np.linalg.eigh(cov_within)
        logging.debug('Cov eig_val= ' +
                      np.array2string(eig_val, precision=3))
        logging.debug('Cov eig_vec=\n' +
                      np.array2string(eig_vec, precision=3))
        eig_val_cum = np.cumsum(eig_val[::-1]) / np.sum(eig_val)
        logging.info(f'One, Two, etc Principal Trait Factors: ' +
                     ', '.join(f'{ev:.1%}'
                               for ev in eig_val_cum) +
                     ', of Variance within Populations')

    logging.shutdown()
